//------------------------------------------------------------------------------
// File: MyCaptureSource.cpp
//
// Desc: Modification of DirectShow's ball source filter that requests bitmaps
//       from the user
//------------------------------------------------------------------------------

#include <streams.h>
#include <olectl.h>
#include <initguid.h>
#include "MyCaptureSource.h"
#include <stdlib.h>
#include <malloc.h>
//#include <stdio.h>

// {BD6EDACD-B236-4324-ABAB-AA6BE46EB933}
DEFINE_GUID(CLSID_MyCaptureSource,
0xbd6edacd, 0xb236, 0x4324, 0xab, 0xab, 0xaa, 0x6b, 0xe4, 0x6e, 0xb9, 0x33);
#pragma warning(disable:4710)  // 'function': function not inlined (optimzation)


void CMyCaptureSource::End()
{
	CAutoLock cAutoLock(&m_cStateLock);
	((CMyCaptureStream*)m_paStreams[0])->End();
	return;
}

void CMyCaptureSource::SetDetails(int type, int width, int height, bool createBuffer)
{
	CAutoLock cAutoLock(&m_cStateLock);
	((CMyCaptureStream*)m_paStreams[0])->SetDetails(type,width,height,createBuffer);
	return;
}

int CMyCaptureSource::GetBufferSize()
{
	CAutoLock cAutoLock(&m_cStateLock);
	return ((CMyCaptureStream*)m_paStreams[0])->GetBufferSize();
}

bool CMyCaptureSource::AddImage(BYTE* pData, int repeatTime)
{
	CAutoLock cAutoLock(&m_cStateLock);
	return ((CMyCaptureStream*)m_paStreams[0])->AddImage(pData,repeatTime);
}

void CMyCaptureStream::End()
{
	CAutoLock cAutoLock(&m_cSharedState);
	
	bWantToEnd=true;
	return;
}

void CMyCaptureStream::SetDetails(int type, int width, int height, bool createBuffer)
{
	CAutoLock cAutoLock(&m_cSharedState);
	
	m_iType        = type;
	m_iImageWidth  = width;
	m_iImageHeight = height;
	bCreateBuffer  = createBuffer;
	return;
}

int CMyCaptureStream::GetBufferSize()
{
	CAutoLock cAutoLock(&m_cSharedState);
	
	return m_sizeToCopy;
}

bool CMyCaptureStream::AddImage(BYTE* pData, int repeatTime)
{
	CAutoLock cAutoLock(&m_cSharedState);

	if(bHasUsed==false)
		return false;
	
	if(pData!=NULL)
	{
		if(bCreateBuffer==true)
		{
			CopyMemory(m_bytes,pData,m_sizeToCopy);
		}
		else
		{
			m_bytes=pData;
		}
		m_iRepeatTime=repeatTime;
		bHasUsed=false;
	}

	return true;
}


HRESULT CMyCaptureStream::DoBufferProcessingLoop(void) {
	
	bool bSleep;
    Command com;
	
    OnThreadStartPlay();
	
    do {
		while (!CheckRequest(&com)) {
			
			{
				CAutoLock cAutoLock(&m_cSharedState);
				if(bWantToEnd==true)
				{
					DeliverEndOfStream();
					return S_OK;
				}
				
				if(bHasUsed==true)
					bSleep=true;
				else
					bSleep=false;
			}
			
			if(bSleep)
			{
				Sleep(1);
				continue;
			}
			
			IMediaSample *pSample;
			
			HRESULT hr = GetDeliveryBuffer(&pSample,NULL,NULL,0);
			if (FAILED(hr)) {
                Sleep(1);
				continue;	// go round again. Perhaps the error will go away
				// or the allocator is decommited & we will be asked to
				// exit soon.
			}
			
			// Virtual function user will override.
			hr = FillBuffer(pSample);
			
			{
				CAutoLock cAutoLock(&m_cSharedState);
				bHasUsed=true;
			}
			
			if (hr == S_OK) {
				hr = Deliver(pSample);
                pSample->Release();
				
                // downstream filter returns S_FALSE if it wants us to
                // stop or an error if it's reporting an error.
                if(hr != S_OK)
                {
					DbgLog((LOG_TRACE, 2, TEXT("Deliver() returned %08x; stopping"), hr));
					return S_OK;
                }
				
			} else if (hr == S_FALSE) {
                // derived class wants us to stop pushing data
				pSample->Release();
				DeliverEndOfStream();
				return S_OK;
			} else {
                // derived class encountered an error
                pSample->Release();
				DbgLog((LOG_ERROR, 1, TEXT("Error %08lX from FillBuffer!!!"), hr));
                DeliverEndOfStream();
                m_pFilter->NotifyEvent(EC_ERRORABORT, hr, 0);
                return hr;
			}
			
            // all paths release the sample
		}
		
        // For all commands sent to us there must be a Reply call!
		
		if (com == CMD_RUN || com == CMD_PAUSE) {
			Reply(NOERROR);
		} else if (com != CMD_STOP) {
			Reply((DWORD) E_UNEXPECTED);
			DbgLog((LOG_ERROR, 1, TEXT("Unexpected command!!!")));
		}
    } while (com != CMD_STOP);
	
    return S_FALSE;
}

// Setup data

const AMOVIESETUP_MEDIATYPE sudOpPinTypes =
{
    &MEDIATYPE_Video,       // Major type
    &MEDIASUBTYPE_NULL      // Minor type
};

const AMOVIESETUP_PIN sudOpPin =
{
    L"Output",              // Pin string name
    FALSE,                  // Is it rendered
    TRUE,                   // Is it an output
    FALSE,                  // Can we have none
    FALSE,                  // Can we have many
    &CLSID_NULL,            // Connects to filter
    NULL,                   // Connects to pin
    1,                      // Number of types
    &sudOpPinTypes };       // Pin details

const AMOVIESETUP_FILTER sudBallax =
{
    &CLSID_MyCaptureSource,    // Filter CLSID
    L"My Capture Source",       // String name
    MERIT_DO_NOT_USE,       // Filter merit
    1,                      // Number pins
    &sudOpPin               // Pin details
};


// COM global table of objects in this dll

CFactoryTemplate g_Templates[] = {
  { L"My Capture Source"
  , &CLSID_MyCaptureSource
  , CMyCaptureSource::CreateInstance
  , NULL
  , &sudBallax }
};
int g_cTemplates = sizeof(g_Templates) / sizeof(g_Templates[0]);


////////////////////////////////////////////////////////////////////////
//
// Exported entry points for registration and unregistration 
// (in this case they only call through to default implementations).
//
////////////////////////////////////////////////////////////////////////

//
// DllRegisterServer
//
// Exported entry points for registration and unregistration
//
STDAPI DllRegisterServer()
{
    return AMovieDllRegisterServer2(TRUE);

} // DllRegisterServer


//
// DllUnregisterServer
//
STDAPI DllUnregisterServer()
{
    return AMovieDllRegisterServer2(FALSE);

} // DllUnregisterServer


//
// DllEntryPoint
//
extern "C" BOOL WINAPI DllEntryPoint(HINSTANCE, ULONG, LPVOID);

BOOL APIENTRY DllMain(HANDLE hModule, 
                      DWORD  dwReason, 
                      LPVOID lpReserved)
{
	return DllEntryPoint((HINSTANCE)(hModule), dwReason, lpReserved);
}

//
// CreateInstance
//
// The only allowed way to create Bouncing balls!
//
CUnknown * WINAPI CMyCaptureSource::CreateInstance(LPUNKNOWN lpunk, HRESULT *phr)
{
    ASSERT(phr);

    CUnknown *punk = new CMyCaptureSource(lpunk, phr);
    if(punk == NULL)
    {
        if(phr)
            *phr = E_OUTOFMEMORY;
    }
    return punk;

} // CreateInstance


//
// Constructor
//
// Initialise a CMyCaptureStream object so that we have a pin.
//
CMyCaptureSource::CMyCaptureSource(LPUNKNOWN lpunk, HRESULT *phr) :
    CSource(NAME("My Capture Source"), lpunk, CLSID_MyCaptureSource)
{
    ASSERT(phr);
    CAutoLock cAutoLock(&m_cStateLock);
	
    m_paStreams = (CSourceStream **) new CMyCaptureStream*[1];
    if(m_paStreams == NULL)
    {
        if(phr)
            *phr = E_OUTOFMEMORY;

        return;
    }

    m_paStreams[0] = new CMyCaptureStream(phr, this, L"Rahil's Capture Source!");
    if(m_paStreams[0] == NULL)
    {
        if(phr)
            *phr = E_OUTOFMEMORY;

        return;
    }

} // (Constructor)


//
// Constructor
//
CMyCaptureStream::CMyCaptureStream(HRESULT *phr,
                         CMyCaptureSource *pParent,
                         LPCWSTR pPinName) :
    CSourceStream(NAME("My Capture Source"),phr, pParent, pPinName),
    m_iImageWidth(300),
    m_iImageHeight(100)
{
	//printf("\nIn Constructor\n");
    ASSERT(phr);
    CAutoLock cAutoLock(&m_cSharedState);

	bHasUsed      = true;
	bWantToEnd    = false;
	bCreateBuffer = true;

	m_bytes=NULL;
	m_iType=5;

} // (Constructor)


//
// Destructor
//
CMyCaptureStream::~CMyCaptureStream()
{
	//printf("In Destructor\n\n");
    CAutoLock cAutoLock(&m_cSharedState);
    if(m_bytes!=NULL && bCreateBuffer==true)
	{
        delete[] m_bytes;
		m_bytes=NULL;
	}

} // (Destructor)


//
// FillBuffer
//
// Plots a ball into the supplied video buffer
//
HRESULT CMyCaptureStream::FillBuffer(IMediaSample *pms)
{
	//printf("In fill buffer\n");
    CheckPointer(pms,E_POINTER);
 

    BYTE *pData;
    long lDataLen;

    pms->GetPointer(&pData);
    lDataLen = pms->GetSize();

    {
        CAutoLock cAutoLockShared(&m_cSharedState);
		
		if(m_bytes==NULL)
		{
			//printf("Null byte pointer\n");
			return E_FAIL;
		}

		ASSERT(m_bytes);
		ASSERT(m_sizeToCopy<=lDataLen);
        // If we haven't just cleared the buffer delete the old
        // ball and move the ball on

		CopyMemory(pData,m_bytes,m_sizeToCopy);

        // The current time is the sample's start
        CRefTime rtStart = m_rtSampleTime;

        // Increment to find the finish time
        m_rtSampleTime += (LONG)m_iRepeatTime;

        pms->SetTime((REFERENCE_TIME *) &rtStart,(REFERENCE_TIME *) &m_rtSampleTime);
    }

    pms->SetSyncPoint(TRUE);
    return NOERROR;

} // FillBuffer


//
// Notify
//
// Alter the repeat rate according to quality management messages sent from
// the downstream filter (often the renderer).  Wind it up or down according
// to the flooding level - also skip forward if we are notified of Late-ness
//
STDMETHODIMP CMyCaptureStream::Notify(IBaseFilter * pSender, Quality q)
{
	return NOERROR;//Ignore
} // Notify


//
// GetMediaType
//
// I _prefer_ 5 formats - 8, 16 (*2), 24 or 32 bits per pixel and
// I will suggest these with an image size of 320x240. However
// I can accept any image size which gives me some space to bounce.
//
// A bit of fun:
//      8 bit displays get red balls
//      16 bit displays get blue
//      24 bit see green
//      And 32 bit see yellow
//
// Prefered types should be ordered by quality, zero as highest quality
// Therefore iPosition =
// 0    return a 32bit mediatype
// 1    return a 24bit mediatype
// 2    return 16bit RGB565
// 3    return a 16bit mediatype (rgb555)
// 4    return 8 bit palettised format
// (iPosition > 4 is invalid)
//
HRESULT CMyCaptureStream::GetMediaType(int iPosition, CMediaType *pmt)
{
	//printf("In get Media Type\n");
    CheckPointer(pmt,E_POINTER);

    CAutoLock cAutoLock(m_pFilter->pStateLock());
    if(iPosition < 0)
    {
        return E_INVALIDARG;
    }

    // Have we run off the end of types?

    if(iPosition > 0)
    {
		//printf("In get Media Type *\n");
        return VFW_S_NO_MORE_ITEMS;
    }

    VIDEOINFO *pvi = (VIDEOINFO *) pmt->AllocFormatBuffer(sizeof(VIDEOINFO));
    if(NULL == pvi)
        return(E_OUTOFMEMORY);

    ZeroMemory(pvi, sizeof(VIDEOINFO));

    switch(m_iType)//iPosition)
    {
        case 7:
        {    
            // Return our highest quality 32bit format

            // since we use RGB888 (the default for 32 bit), there is
            // no reason to use BI_BITFIELDS to specify the RGB
            // masks. Also, not everything supports BI_BITFIELDS

            SetPaletteEntries(Yellow);
            pvi->bmiHeader.biCompression = BI_RGB;
            pvi->bmiHeader.biBitCount    = 32;
            break;
        }

        case 6:
        {   // Return our 24bit format

            SetPaletteEntries(Green);
            pvi->bmiHeader.biCompression = BI_RGB;
            pvi->bmiHeader.biBitCount    = 24;
            break;
        }

        case 5:
        {       
            // 16 bit per pixel RGB565

            // Place the RGB masks as the first 3 doublewords in the palette area
            for(int i = 0; i < 3; i++)
                pvi->TrueColorInfo.dwBitMasks[i] = bits565[i];

            SetPaletteEntries(Blue);
            pvi->bmiHeader.biCompression = BI_BITFIELDS;
            pvi->bmiHeader.biBitCount    = 16;
            break;
        }

        case 4:
        {   // 16 bits per pixel RGB555

            // Place the RGB masks as the first 3 doublewords in the palette area
            for(int i = 0; i < 3; i++)
                pvi->TrueColorInfo.dwBitMasks[i] = bits555[i];

            SetPaletteEntries(Blue);
            pvi->bmiHeader.biCompression = BI_BITFIELDS;
            pvi->bmiHeader.biBitCount    = 16;
            break;
        }

        case 3:
        {   // 16 bits per pixel RGB444

            // Place the RGB masks as the first 3 doublewords in the palette area
            pvi->TrueColorInfo.dwBitMasks[0] = 0x000F;
			pvi->TrueColorInfo.dwBitMasks[1] = 0x00F0;
			pvi->TrueColorInfo.dwBitMasks[2] = 0x0F00;

            SetPaletteEntries(Blue);
            pvi->bmiHeader.biCompression = BI_BITFIELDS;
            pvi->bmiHeader.biBitCount    = 16;
            break;
        }

        case 2:
        {   // 8 bit palettised

            SetPaletteEntries(Red);
            pvi->bmiHeader.biCompression = BI_RGB;
            pvi->bmiHeader.biBitCount    = 8;
            pvi->bmiHeader.biClrUsed     = iPALETTE_COLORS;
            break;
        }

		case 1:
        {   // 4 bit palettised

            SetPaletteEntries(Red);
            pvi->bmiHeader.biCompression = BI_RGB;
            pvi->bmiHeader.biBitCount    = 4;
            pvi->bmiHeader.biClrUsed     = iPALETTE_COLORS;
            break;
        }

		case 0:
        {   // 1 bit palettised

            SetPaletteEntries(Red);
            pvi->bmiHeader.biCompression = BI_RGB;
            pvi->bmiHeader.biBitCount    = 1;
            pvi->bmiHeader.biClrUsed     = iPALETTE_COLORS;
            break;
        }
    }

    // (Adjust the parameters common to all formats...)

    // put the optimal palette in place
    for(int i = 0; i < iPALETTE_COLORS; i++)
    {
        pvi->TrueColorInfo.bmiColors[i].rgbRed      = m_Palette[i].peRed;
        pvi->TrueColorInfo.bmiColors[i].rgbBlue     = m_Palette[i].peBlue;
        pvi->TrueColorInfo.bmiColors[i].rgbGreen    = m_Palette[i].peGreen;
        pvi->TrueColorInfo.bmiColors[i].rgbReserved = 0;
    }

    pvi->bmiHeader.biSize       = sizeof(BITMAPINFOHEADER);
    pvi->bmiHeader.biWidth      = m_iImageWidth;
    pvi->bmiHeader.biHeight     = m_iImageHeight;
    pvi->bmiHeader.biPlanes     = 1;
    pvi->bmiHeader.biSizeImage  = GetBitmapSize(&pvi->bmiHeader);
    pvi->bmiHeader.biClrImportant = 0;

    SetRectEmpty(&(pvi->rcSource)); // we want the whole image area rendered.
    SetRectEmpty(&(pvi->rcTarget)); // no particular destination rectangle

    pmt->SetType(&MEDIATYPE_Video);
    pmt->SetFormatType(&FORMAT_VideoInfo);
    pmt->SetTemporalCompression(FALSE);

    // Work out the GUID for the subtype from the header info.
    const GUID SubTypeGUID = GetBitmapSubtype(&pvi->bmiHeader);
    pmt->SetSubtype(&SubTypeGUID);
    pmt->SetSampleSize(pvi->bmiHeader.biSizeImage);

    return NOERROR;

} // GetMediaType


//
// CheckMediaType
//
// We will accept 8, 16, 24 or 32 bit video formats, in any
// image size that gives room to bounce.
// Returns E_INVALIDARG if the mediatype is not acceptable
//
HRESULT CMyCaptureStream::CheckMediaType(const CMediaType *pMediaType)
{
	//printf("Checking Media Type\n");
    CheckPointer(pMediaType,E_POINTER);

    if((*(pMediaType->Type()) != MEDIATYPE_Video) ||   // we only output video
       !(pMediaType->IsFixedSize()))                   // in fixed size samples
    {
		//printf(" Error Not fixed size or not video type\n");
        return E_INVALIDARG;
    }

    // Check for the subtypes we support
    const GUID *SubType = pMediaType->Subtype();
    if (SubType == NULL)
	{
		//printf(" Subtype is NULL\n");
        return E_INVALIDARG;
	}

    if(!(
		     (*SubType == MEDIASUBTYPE_RGB1     && m_iType==0)
		  || (*SubType == MEDIASUBTYPE_RGB4     && m_iType==1)
		  || (*SubType == MEDIASUBTYPE_RGB8     && m_iType==2)
		  || (*SubType == MEDIASUBTYPE_ARGB4444 && m_iType==3)
		  || (*SubType == MEDIASUBTYPE_RGB555   && m_iType==4)
		  || (*SubType == MEDIASUBTYPE_RGB565   && m_iType==5)
		  || (*SubType == MEDIASUBTYPE_RGB24    && m_iType==6)
		  || (*SubType == MEDIASUBTYPE_RGB32    && m_iType==7)
	  ))
    {
		//printf(" Subtype and m_iType *%i* do not match\n",m_iType);
        return E_INVALIDARG;
    }

    // Get the format area of the media type
    VIDEOINFO *pvi = (VIDEOINFO *) pMediaType->Format();

    if(pvi == NULL)
	{
		//printf(" Videoinfo is NULL\n");
        return E_INVALIDARG;
	}

    // Check if the image width & height have changed
    if(pvi->bmiHeader.biWidth != m_iImageWidth || 
       pvi->bmiHeader.biHeight != m_iImageHeight)
    {
		//printf(" Sizes changed\n");
        // If the image width/height is changed, fail CheckMediaType() to force
        // the renderer to resize the image.
        return E_INVALIDARG;
    }


    return S_OK;  // This format is acceptable.

} // CheckMediaType


//
// DecideBufferSize
//
// This will always be called after the format has been sucessfully
// negotiated. So we have a look at m_mt to see what size image we agreed.
// Then we can ask for buffers of the correct size to contain them.
//
HRESULT CMyCaptureStream::DecideBufferSize(IMemAllocator *pAlloc,
                                      ALLOCATOR_PROPERTIES *pProperties)
{
	//printf("\tDeciding Buffer Size\n");
    CheckPointer(pAlloc,E_POINTER);
    CheckPointer(pProperties,E_POINTER);

    CAutoLock cAutoLock(m_pFilter->pStateLock());
    HRESULT hr = NOERROR;

    VIDEOINFO *pvi = (VIDEOINFO *) m_mt.Format();
    pProperties->cBuffers = 1;
    pProperties->cbBuffer = pvi->bmiHeader.biSizeImage;
	pProperties->cbAlign  = 1;
	pProperties->cbPrefix = 0;

    ASSERT(pProperties->cbBuffer);

    // Ask the allocator to reserve us some sample memory, NOTE the function
    // can succeed (that is return NOERROR) but still not have allocated the
    // memory that we requested, so we must check we got whatever we wanted

    ALLOCATOR_PROPERTIES Actual;
    hr = pAlloc->SetProperties(pProperties,&Actual);
    if(FAILED(hr))
    {
		//printf("\t Failed to set properties\n");
        return hr;
    }

    // Is this allocator unsuitable

    if(Actual.cbBuffer < pProperties->cbBuffer)
    {
		//printf("\t Actual buffer size is different\n");
        return E_FAIL;
    }

    // Make sure that we have only 1 buffer (we erase the ball in the
    // old buffer to save having to zero a 200k+ buffer every time
    // we draw a frame)

    ASSERT(Actual.cBuffers == pProperties->cBuffers);
    return NOERROR;

} // DecideBufferSize


//
// SetMediaType
//
// Called when a media type is agreed between filters
//
HRESULT CMyCaptureStream::SetMediaType(const CMediaType *pMediaType)
{
	//printf("In set media type\n");
    CAutoLock cAutoLock(m_pFilter->pStateLock());

    // Pass the call up to my base class

    HRESULT hr = CSourceStream::SetMediaType(pMediaType);

    if(SUCCEEDED(hr))
    {
        VIDEOINFO * pvi = (VIDEOINFO *) m_mt.Format();
        if (pvi == NULL)
            return E_UNEXPECTED;

		m_sizeToCopy = pvi->bmiHeader.biSizeImage;

		if(bCreateBuffer==true)
		{
			if(m_bytes!=NULL)
			{
				delete[] m_bytes;
				m_bytes=NULL;
			}

			m_bytes = new BYTE[m_sizeToCopy];
			if(m_bytes==NULL)
			{
				//printf("\t UhOh\n");
				return E_OUTOFMEMORY;
			}
		}

		//printf("\t OK %p %i\n",m_bytes,m_sizeToCopy);

        return NOERROR;
    } 

    return hr;

} // SetMediaType


//
// OnThreadCreate
//
// As we go active reset the stream time to zero
//
HRESULT CMyCaptureStream::OnThreadCreate()
{
    CAutoLock cAutoLockShared(&m_cSharedState);
    m_rtSampleTime = 0;

    // we need to also reset the repeat time in case the system
    // clock is turned off after m_iRepeatTime gets very big
    m_iRepeatTime = 0;

    return NOERROR;

} // OnThreadCreate


//
// SetPaletteEntries
//
// If we set our palette to the current system palette + the colours we want
// the system has the least amount of work to do whilst plotting our images,
// if this stream is rendered to the current display. The first non reserved
// palette slot is at m_Palette[10], so put our first colour there. Also
// guarantees that black is always represented by zero in the frame buffer
//
HRESULT CMyCaptureStream::SetPaletteEntries(Colour color)
{
    CAutoLock cAutoLock(m_pFilter->pStateLock());

    HDC hdc = GetDC(NULL);  // hdc for the current display.
    UINT res = GetSystemPaletteEntries(hdc, 0, iPALETTE_COLORS, (LPPALETTEENTRY) &m_Palette);
    ReleaseDC(NULL, hdc);

    if(res == 0)
        return E_FAIL;

    switch(color)
    {
        case Red:
            m_Palette[10].peBlue  = 0;
            m_Palette[10].peGreen = 0;
            m_Palette[10].peRed   = 0xff;
            break;

        case Yellow:
            m_Palette[10].peBlue  = 0;
            m_Palette[10].peGreen = 0xff;
            m_Palette[10].peRed   = 0xff;
            break;

        case Blue:
            m_Palette[10].peBlue  = 0xff;
            m_Palette[10].peGreen = 0;
            m_Palette[10].peRed   = 0;
            break;

        case Green:
            m_Palette[10].peBlue  = 0;
            m_Palette[10].peGreen = 0xff;
            m_Palette[10].peRed   = 0;
            break;
    }

    m_Palette[10].peFlags = 0;
    return NOERROR;

} // SetPaletteEntries


