#include "MotionData.h"
#include "3d_object.h"
#include "ContactInfo.h"
#include "LinearProgramSolver.h"
#include <stdio.h>


double     Global_T                  = 0.1;  //Time Step
double     Global_UnitsPerMetre      = 0.2;  //Use this to scale the effects of gravity
double     Global_VelocityZero       = 0.2;  //Threshold velocity used in deciding which contacts are resting contacts
bool       bUseMinForce              = true; //Used by qDash, if this is set to false it will try out all combinations
                                             //   of vanishing contact points, otherwise it will approximate the
									         //   solution using the min force.
CDiffEqMotionData* mdArray           = NULL; //An array of motion data, one for each object we are simulating
CDiffEqMotionData* mdArrayDot        = NULL; //The rate of change of each of the members of mdArray;
CContact*          contactArray      = NULL; //Array of possible contact points
int*               indexContactArray = NULL; //Array of resting contact points


//Constructor of CMotionData
//Just sets some sensible default values for the properties
CMotionData::CMotionData()
{
	Orientation.Identity();
	Position = CVector(0,0,0);

	Inertia = 1;

	new_Orientation     = old_Orientation;
	new_Position        = old_Position;
	new_AngularMomentum = old_AngularMomentum;
	new_Velocity        = old_Velocity;
}



int InitGlobalVariables(int numOfObjects)
{
	mdArray              = NULL;
	mdArrayDot           = NULL;
	contactArray         = NULL;
	indexContactArray    = NULL;

	mdArray = new CDiffEqMotionData[numOfObjects];
	if(mdArray==NULL)
	{
		ReleaseGlobalVariables();
		return 0;
	}

	mdArrayDot = new CDiffEqMotionData[numOfObjects];
	if(mdArrayDot==NULL)
	{
		ReleaseGlobalVariables();
		return 0;
	}

	contactArray = new CContact[4*numOfObjects*(numOfObjects-1)];
	if(contactArray==NULL)
	{
		ReleaseGlobalVariables();
		return 0;
	}

	indexContactArray = new int[4*numOfObjects*(numOfObjects-1)];
	if(contactArray==NULL)
	{
		ReleaseGlobalVariables();
		return 0;
	}

	return 1;
}

void ReleaseGlobalVariables()
{
	if(mdArray!=NULL)
		delete[] mdArray;
	
	if(mdArrayDot!=NULL)
		delete[] mdArrayDot;

	if(contactArray!=NULL)
		delete[] contactArray;

	if(indexContactArray!=NULL)
		delete[] indexContactArray;

	mdArray              = NULL;
	mdArrayDot           = NULL;
	contactArray         = NULL;
	indexContactArray    = NULL;

	return;
}









//Used to convert the states of the object into an array which can then be integrated
void MotionDataToArray(double* array, CDiffEqMotionData* pMD, int NoOfObjects)
{
	int i;
	for(i=0;i<NoOfObjects;i++)
	{
		array[i*10+0] = pMD[i].Mass;
		array[i*10+1] = pMD[i].Inertia;
		array[i*10+2] = pMD[i].Orientation;
		array[i*10+3] = pMD[i].Position.e[0];
		array[i*10+4] = pMD[i].Position.e[1];
		array[i*10+5] = pMD[i].Position.e[2];
		array[i*10+6] = pMD[i].AngularMomentum;
		array[i*10+7] = pMD[i].Velocity.e[0];
		array[i*10+8] = pMD[i].Velocity.e[1];
		array[i*10+9] = pMD[i].Velocity.e[2];
	}
	
	return;
}

//Used to convert the output from integrating motion back into the states of the objects
void ArrayToMotionData(double* array, CDiffEqMotionData* pMD, int NoOfObjects)
{
	int i;
	for(i=0;i<NoOfObjects;i++)
	{
		pMD[i].Mass            = array[i*10+0];
		pMD[i].Inertia         = array[i*10+1];
		pMD[i].Orientation     = array[i*10+2];
		pMD[i].Position.e[0]   = array[i*10+3];
		pMD[i].Position.e[1]   = array[i*10+4];
		pMD[i].Position.e[2]   = array[i*10+5];
		pMD[i].AngularMomentum = array[i*10+6];
		pMD[i].Velocity.e[0]   = array[i*10+7];
		pMD[i].Velocity.e[1]   = array[i*10+8];
		pMD[i].Velocity.e[2]   = array[i*10+9];
	}
	
	return;
}



//This function gives the equations of motion that needed to be integrated
void qDash(CVectorList* x_dot, CVectorList* x, double t)
{
	int i,j;
	int NoOfObjects;
	int NoOfRestingContacts;
	CVector g;

	//Set acceleration due to gravity
	g = CVector(0,0,-9.8*Global_UnitsPerMetre);

	//Work out number of objects we are integrating
	NoOfObjects = x->dimension/10;

	//Just in case
	x_dot->Fill(0);
	for(i=0;i<NoOfObjects;i++)
	{
		mdArrayDot[i].Mass            = 0;
		mdArrayDot[i].Inertia         = 0;
		mdArrayDot[i].Orientation     = 0;
		mdArrayDot[i].Position        = CVector(0,0,0);
		mdArrayDot[i].AngularMomentum = 0;
		mdArrayDot[i].Velocity        = CVector(0,0,0);
	}

	//Fill in state of simulation from the variable x
	ArrayToMotionData(x->array,mdArray,NoOfObjects);

	//Work out number of resting contact points
	NoOfRestingContacts = 0;
	for(i=0;i<4*NoOfObjects*(NoOfObjects-1);i++)
	{
		if(contactArray[i].bContactExists==true && contactArray[i].Extra.bSmallVelocityContact==true)
		{		
			//Check that contact is not between two infinite masses
			if(mdArray[contactArray[i].idObjectA].Mass>0 || mdArray[contactArray[i].idObjectB].Mass>0)
			{
				indexContactArray[NoOfRestingContacts] = i;
				NoOfRestingContacts++;
			}
		}
	}

	//NoOfRestingContacts is now equal to the number of unknown Forces we need to calculate
	//Lets call it F for easier to read code
	int F = NoOfRestingContacts;

	//***Find the forces***
	CMatrixList mF,mLP,mMinF;

	if(F>0)
	{
		//mF is a list of the magnitudes of the unknown forces
		if(!mF.setDimensions(F,1))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		if(!mMinF.setDimensions(F,1))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		mF.Fill(0);

		//mLP is the matrix describing the linear program
		if(!mLP.setDimensions(2*F+1,F+2))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		//Zero out
		mLP.Fill(0);

		//We are trying to find a feasible solution
		//So the first row is should be zero, which it is
		
		//Forces>=0
		for(i=1;i<=F;i++)
		{
			mLP.SetValue(F+1+i,  i, 1);
			mLP.SetValue(F+1+i,F+1,-1);
		}

		//Fill in the acceleration constraints
		for(i=1;i<=F;i++)
		for(j=1;j<=F;j++)
		{
			int     Ai,Bi;
			int     Aj,Bj;
			CVector ri,ni;
			CVector rj,nj;
			CMatrix R;
			double  result;
			double  TotalResult;
			
			TotalResult = 0;

			Ai = contactArray[indexContactArray[i-1]].idObjectA;
			Bi = contactArray[indexContactArray[i-1]].idObjectB;
			Aj = contactArray[indexContactArray[j-1]].idObjectA;
			Bj = contactArray[indexContactArray[j-1]].idObjectB;

			ri = contactArray[indexContactArray[i-1]].objectA_IntersectionPoint;
			ni = contactArray[indexContactArray[i-1]].objectB_OutwardNormal;
			rj = contactArray[indexContactArray[j-1]].objectA_IntersectionPoint;
			nj = contactArray[indexContactArray[j-1]].objectB_OutwardNormal;

			//Transform To World Coordinates
			R.Rotation(0,1,0,mdArray[Ai].Orientation);
			ri = R*ri+mdArray[Ai].Position;
			R.Rotation(0,1,0,mdArray[Bi].Orientation);
			ni = R*ni;
			R.Rotation(0,1,0,mdArray[Aj].Orientation);
			rj = R*rj+mdArray[Aj].Position;
			R.Rotation(0,1,0,mdArray[Bj].Orientation);
			nj = R*nj;
			

			if((Aj==Ai || Aj==Bi) && mdArray[Aj].Mass>0)
			{
				result  = 0;
				result += ni.Dot(nj)/mdArray[Aj].Mass;
				result += (((ri-mdArray[Aj].Position).Cross(ni)).Cross(rj-mdArray[Aj].Position)).Dot(nj)
					    / (mdArray[Aj].Inertia*mdArray[Aj].Mass);

				if(Aj==Ai)
					TotalResult += result;

				if(Aj==Bi)
					TotalResult -= result;
			}

			if((Bj==Ai || Bj==Bi) && mdArray[Bj].Mass>0)
			{
				result  = 0;
				result += ni.Dot(nj)/mdArray[Bj].Mass;
				result += (((ri-mdArray[Bj].Position).Cross(ni)).Cross(rj-mdArray[Bj].Position)).Dot(nj)
					    / (mdArray[Bj].Inertia*mdArray[Bj].Mass);

				if(Bj==Ai)
					TotalResult -= result;

				if(Bj==Bi)
					TotalResult += result;
			}

			mLP.SetValue(j+1,i,TotalResult);
		}

		//Fill in the constant part of the acceleration constraints
		for(i=1;i<=F;i++)
		{
			//Fill in >= sign
			mLP.SetValue(i+1,F+1,-1);

			int     A,B;
			CVector r,n;
			CVector wA,wB;
			CMatrix R;
			double  result;
			
			result = 0;

			A = contactArray[indexContactArray[i-1]].idObjectA;
			B = contactArray[indexContactArray[i-1]].idObjectB;

			r = contactArray[indexContactArray[i-1]].objectA_IntersectionPoint;
			n = contactArray[indexContactArray[i-1]].objectB_OutwardNormal;

			//Transform To World Coordinates
			R.Rotation(0,1,0,mdArray[A].Orientation);
			r = R*r+mdArray[A].Position;
			R.Rotation(0,1,0,mdArray[B].Orientation);
			n = R*n;

			wA = CVector(0,1,0)*mdArray[A].AngularMomentum/mdArray[A].Inertia;
			wB = CVector(0,1,0)*mdArray[B].AngularMomentum/mdArray[B].Inertia;

			if(mdArray[A].Mass>0 && mdArray[B].Mass>0)
			{
				result  = 0;
				result += (mdArray[A].Velocity+wA.Cross(r-mdArray[A].Position)
					      -mdArray[B].Velocity-wB.Cross(r-mdArray[B].Position)
						  ).Dot(wB.Cross(n))*2;
				result -= ((wA.Dot(wA))*(r-mdArray[A].Position)
					      -(wB.Dot(wB))*(r-mdArray[B].Position)
						  ).Dot(n);
			}

			if(mdArray[A].Mass<=0)
			{
				result  = 0;
				result -= (mdArray[B].Velocity+wB.Cross(r-mdArray[B].Position)).Dot(wB.Cross(n))*2;
				result += (wB.Dot(wB))*(r-mdArray[B].Position).Dot(n);
				result -= g.Dot(n);
			}

			if(mdArray[B].Mass<=0)
			{
				result  = 0;
				result -= (wA.Dot(wA))*(r-mdArray[A].Position).Dot(n);
				result += g.Dot(n);
			}

			mLP.SetValue(i+1,F+2,-result);
		}

		//mLP completely filled in

		if(bUseMinForce==true)
		{
			for(i=1;i<=2*F+1;i++)
				mLP.SetValue(i,F+1,-1);
			
			for(i=1;i<=F;i++)
				mLP.SetValue(1,i,1);

			//***Find Min Force***
			int returnResult;
			returnResult = glpkSimplexAlgorithm(mLP,mMinF);
			
			if(returnResult==-1)
			{
				printf("*** Ran Out Of Memory ***\n");
				return;
			}

			//UhOh
			if(returnResult!=1)
				mMinF.Fill(0);

			switch(returnResult)
			{
			case 1:
				//All went well
				break;

			case -2:
				//Should never occur
				printf("No Feasible Solution\n");
				break;
				
			case -3:
				//Should never occur
				printf("Solution Is Unbounded\n");
				break;
				
			case -4:
				//Should never occur
				printf("Formulation Of Linear Program Is Too Small\n");
				break;
				
			case -5:
				//Should never occur
				printf("No Of Iterations Excessive\n");
				break;
				
			default:
				//Should never occur
				printf("Error Solver Returned %i\n",returnResult);
			}

			//Use mMinF to choose vanising points
			mLP.SetValue(1,F+1,0);

			for(j=1;j<=F;j++)
			{
				double result;
				result = -mLP.GetValue(j+1,F+2);
				for(i=1;i<=F;i++)
					result+=mLP.GetValue(j+1,i)*mMinF.GetValue(i,1);

				if(result>=0.0001)//>0
				{
					//Guess contact j is vanishing
					mLP.SetValue(j+1,F+1,-1);
					mLP.SetValue(F+j+1,F+1,0);
				}
				else
				{
					//Guess contact j is non-vanishing
					mLP.SetValue(j+1,F+1,0);
					mLP.SetValue(F+j+1,F+1,-1);
				}
			}

			//Find mF
			returnResult = glpkSimplexAlgorithm(mLP,mF);
			
			if(returnResult==-1)
			{
				printf("*** Ran Out Of Memory ***\n");
				return;
			}

			switch(returnResult)
			{
			case 1:
				//All went well
				break;

			case -2:
				//We made a bad choice
				//Just use mMinF
				mF = mMinF;
				
				mF.getDimensions(&j,&i);
				if(i<=0 || j<=0)
				{
					printf("*** Ran Out Of Memory ***\n");
					return;
				}

				break;

			case -3:
				//Should never occur
				printf("Solution Is Unbounded\n");
				break;
				
			case -4:
				//Should never occur
				printf("Formulation Of Linear Program Is Too Small\n");
				break;
				
			case -5:
				//Should never occur
				printf("No Of Iterations Excessive\n");
				break;
				
			default:
				//Should never occur
				printf("Error Solver Returned %i\n",returnResult);
			}

		}//End of use Min Force
		else
		{
			//***Cycle Through Choices Of Vanishing Points***
			int NoOfChoices=1;
			for(i=1;i<=F;i++)
				NoOfChoices*=2;
			
			for(i=0;i<=NoOfChoices-1;i++)
			{
				//Set elements in mF that will be zero
				int temp = i;
				for(j=1;j<=F;j++)
				{
					if(temp%2==0)
					{
						mLP.SetValue(F+1+j,F+1,-1);
						mLP.SetValue(j+1,F+1,0);
					}
					else
					{
						mLP.SetValue(F+1+j,F+1,0);
						mLP.SetValue(j+1,F+1,-1);
					}
					
					temp/=2;
				}
				
				//Find Forces
				int returnResult;
				returnResult = SimplexAlgorithm(mLP,mF);
				
				//success
				if(returnResult==1)
					break;
				
				if(returnResult==-1)
				{
					printf("*** Ran Out Of Memory ***\n");
					return;
				}
				
				switch(returnResult)
				{
				case -2:
					//Should occur rarely
					//printf("No Feasible Solution %i %i\n",F, i);
					if(i==NoOfChoices-1)
						//Should never occur
						printf("*** ERROR: No Choice Of Set Was Feasible***\n");
					break;
					
				case -3:
					//Should never occur
					printf("Solution Is Unbounded\n");
					break;
					
				case -4:
					//Should never occur
					printf("Formulation Of Linear Program Is Too Small\n");
					break;
					
				case -5:
					//Should never occur
					printf("No Of Iterations Excessive\n");
					break;
					
				default:
					//Should never occur
					printf("Error Solver Returned %i\n",returnResult);
				}
			}//Next Choice
		}//End of don't use Min Force determination
	}
	//Found Forces


	//***Update mdArrayDot[]***


	//Add gravity
	for(i=0;i<NoOfObjects;i++)
	{
		if(mdArray[i].Mass>0)
		{
			mdArrayDot[i].Orientation     = mdArray[i].AngularMomentum/mdArray[i].Inertia;
			mdArrayDot[i].Position        = mdArray[i].Velocity;
			mdArrayDot[i].AngularMomentum = 0;
			mdArrayDot[i].Velocity        = g;
		}
	}

	//Add effect of previously unknown forces
	for(i=0;i<NoOfRestingContacts;i++)
	{
		int A,B;
		CVector r,n,Torque;
		CMatrix R;

		A = contactArray[indexContactArray[i]].idObjectA;
		B = contactArray[indexContactArray[i]].idObjectB;

		r = contactArray[indexContactArray[i]].objectA_IntersectionPoint;
		n = contactArray[indexContactArray[i]].objectB_OutwardNormal;

		//change to matrix
		R.Rotation(0,1,0,mdArray[A].Orientation);
		r = R*r+mdArray[A].Position;
		R.Rotation(0,1,0,mdArray[B].Orientation);
		n = R*n;

		//Object A
		if(mdArray[A].Mass>0)
		{
			Torque = (r - mdArray[A].Position).Cross(n)*mF.GetValue(i+1,1);

			mdArrayDot[A].AngularMomentum += Torque.e[1]/mdArray[A].Mass;
			mdArrayDot[A].Velocity        += n*mF.GetValue(i+1,1)/mdArray[A].Mass;
		}

		//Object B
		if(mdArray[B].Mass>0)
		{
			Torque = (r - mdArray[B].Position).Cross(n)*mF.GetValue(i+1,1);

			mdArrayDot[B].AngularMomentum -= Torque.e[1]/mdArray[B].Mass;
			mdArrayDot[B].Velocity        -= n*mF.GetValue(i+1,1)/mdArray[B].Mass;
		}
	}


	MotionDataToArray(x_dot->array,mdArrayDot,NoOfObjects);
	return;
}


//Assumes objects are orientated with respect to axis and centered about the origin
//Ignores y coordinates
//Array should be of size 4n(n-1) where n is the number of boxes in the simulation
//The array is filled in so all entries with the same idObjectA, idObjectB
//are next to each other
void InitializeCContactArray(C3dObject* pMesh, int NoOfObjects, CContact* Array)
{
	double tolerance = 0.01;
	int i,j,k,a,b;
	int counter;
	int index;

	for(i=0;i<NoOfObjects;i++)
	{
		for(a=-1;a<=1;a+=2)
		for(b=-1;b<=1;b+=2)
		{
			//a,b specify which quadrant to search
			//we assume y coordinate is +ve

			for(k=0;k<pMesh[i].iNumOfVerts;k++)
			{
				//Check all vertices till we find one in the right area

				if(pMesh[i].vertList[k].y>tolerance
				&& pMesh[i].vertList[k].x*a>tolerance
				&& pMesh[i].vertList[k].z*b>tolerance)
				{
					counter = 0;
					for(j=0;j<NoOfObjects;j++)
					{
						if(i==j)
							continue;

						//How many have already been filled
						index = i*4*(NoOfObjects-1);
						//Offset for vertex
						index += (a+1)+(b+1)/2;
						//Offset for object
						index += counter*4;

						Array[index].bContactExists                 = false;
						Array[index].idObjectA                      = i;
						Array[index].idObjectB                      = j;
						Array[index].objectA_IntersectionPoint.e[0] = pMesh[i].vertList[k].x;
						Array[index].objectA_IntersectionPoint.e[1] = 0;
						Array[index].objectA_IntersectionPoint.e[2] = pMesh[i].vertList[k].z;
						Array[index].Extra.bNormalHoldsInfo         = false;
						Array[index].Extra.bSmallVelocityContact    = false;
						counter++;
					}

					break;
				}

				//As soon as we found one we stop searching
			}

			//Find New Corner

		}

		//New Object To Find Vertices Of
	}

	return;
}



//Assumes orientated wrt to axis and centered about Origin
//Ignores y coordinate
//Choosing the correct collision normals makes this function quite complicated
void TestCuboidIntersection(C3dObject* pMesh, int NoOfObjects, CContact* Array)
{
	int i,j;
	int A,B;
	CVector Max,v;
	CMatrix oldO;
	CVector oldP;

	//Used to define a distance that is close to a corner
	//Should be slightly larger than 1.5*maximum penetration expected
	double cornerDistance = 5;

	for(i=0;i<NoOfObjects*(NoOfObjects-1);i++)
	{
		//All have the same Object A,B ids but different indexOfIntersectionPoint
		A = Array[i*4].idObjectA;
		B = Array[i*4].idObjectB;


		Max.e[0] = pMesh[B].vertList[0].x;
		Max.e[1] = pMesh[B].vertList[0].y;
		Max.e[2] = pMesh[B].vertList[0].z;
		
		for(j=0;j<3;j++)
		{
			if(Max.e[j]<0)
				Max.e[j]*=-1;
		}
		
		oldO = pMesh[A].data.Orientation;
		oldP = pMesh[A].data.Position;
		
		pMesh[A].data.Orientation = pMesh[B].data.Orientation.ReturnInverse()*oldO;
		pMesh[A].data.Position    = pMesh[B].data.Orientation.ReturnInverse()*(oldP-pMesh[B].data.Position);
		
		//Test Object A's 4 vertices for intersection in Object B
		for(j=0;j<4;j++)
		{
			v      = pMesh[A].data.Orientation*Array[i*4+j].objectA_IntersectionPoint+pMesh[A].data.Position;
			v.e[1] = 0;

			if(-Max.e[0]<=v.e[0] && v.e[0]<=Max.e[0]
			&& -Max.e[2]<=v.e[2] && v.e[2]<=Max.e[2])
			{
				//Vertex of ObjA is inside ObjB
				Array[i*4+j].bContactExists = true;

				//***Decide on normal of collision***

				//A very good guess would be the normal of the closest edge			
				double CurrentDistance;
				double ClosestDistance;
				
				//Top Edge
				CurrentDistance = Max.e[2]-v.e[2];
				ClosestDistance = CurrentDistance;
				Array[i*4+j].objectB_OutwardNormal = CVector(0,0,1);
				
				//Bottom Edge
				CurrentDistance = Max.e[2]+v.e[2];
				if(ClosestDistance>CurrentDistance)
				{
					ClosestDistance = CurrentDistance;
					Array[i*4+j].objectB_OutwardNormal = CVector(0,0,-1);
				}
				
				//Left Edge
				CurrentDistance = Max.e[0]+v.e[0];
				if(ClosestDistance>CurrentDistance)
				{
					ClosestDistance = CurrentDistance;
					Array[i*4+j].objectB_OutwardNormal = CVector(-1,0,0);
				}
				
				//Right Edge
				CurrentDistance = Max.e[0]-v.e[0];
				if(ClosestDistance>CurrentDistance)
				{
					ClosestDistance = CurrentDistance;
					Array[i*4+j].objectB_OutwardNormal = CVector(1,0,0);
				}
				//End of finding closest edge

				//If the point is close to a corner then the closest edge
				//approach is not good enough
				bool    bCloseToACorner;
				CVector HChoice, VChoice;
				CVector Corner;
				double  radius;

				//If the point is less the cornerDistance away from a corner
				//we will say it is close to a corner
				radius          = cornerDistance;
				bCloseToACorner = false;

				//Top Left Corner
				Corner = CVector(-Max.e[0],0,Max.e[2]);
				if((Corner-v).Length()<radius)
				{
					bCloseToACorner = true;
					HChoice = CVector(-1,0,0);
					VChoice = CVector(0,0,1);
				}
				
				//Top Right Corner
				Corner = CVector(Max.e[0],0,Max.e[2]);
				if((Corner-v).Length()<radius)
				{
					bCloseToACorner = true;
					HChoice = CVector(1,0,0);
					VChoice = CVector(0,0,1);
				}
				
				//Bottom Left Corner
				Corner = CVector(-Max.e[0],0,-Max.e[2]);
				if((Corner-v).Length()<radius)
				{
					bCloseToACorner = true;
					HChoice = CVector(-1,0,0);
					VChoice = CVector(0,0,-1);
				}
				
				//Bottom Right Corner
				Corner = CVector(Max.e[0],0,-Max.e[2]);
				if((Corner-v).Length()<radius)
				{
					bCloseToACorner = true;
					HChoice = CVector(1,0,0);
					VChoice = CVector(0,0,-1);
				}

				if(bCloseToACorner==true)
				{
					//Best guess is the choice that's closest to a 
					//a vector that bisects the angle at the point of A
					//and points in the direction of the centre of A
					CVector N;
					CVector u;

					u = Array[i*4+j].objectA_IntersectionPoint;

					//Top Left Corner
					if(u.e[0]<0 && u.e[2]>0)
						N = CVector(0.5,0,-0.5);

					//Top Right Corner
					if(u.e[0]>0 && u.e[2]>0)
						N = CVector(-0.5,0,-0.5);

					//Bottom Left Corner
					if(u.e[0]<0 && u.e[2]<0)
						N = CVector(0.5,0,0.5);

					//Bottom Right Corner
					if(u.e[0]>0 && u.e[2]<0)
						N = CVector(-0.5,0,0.5);

					N = pMesh[A].data.Orientation*N;

					if(N.Dot(HChoice)>N.Dot(VChoice))
						Array[i*4+j].objectB_OutwardNormal = HChoice;
					else
						Array[i*4+j].objectB_OutwardNormal = VChoice;

					//A good guess is the last normal
					if(Array[i*4+j].Extra.bNormalHoldsInfo==true)
					{
						//Check if the previous normal is HChoice or VChoice
						if((HChoice-v).Length()<0.5 || (VChoice-v).Length())
							Array[i*4+j].objectB_OutwardNormal = Array[i*4+j].Extra.old_Normal;
					}
				}	


				Array[i*4+j].Extra.bNormalHoldsInfo = true;
				Array[i*4+j].Extra.old_Normal       = Array[i*4+j].objectB_OutwardNormal;

				//End of vertex of ObjA is inside ObjB
			}
			else
			{
				//Vertex of ObjA is outside ObjB
				Array[i*4+j].bContactExists         = false;
				Array[i*4+j].Extra.bNormalHoldsInfo = false;
			}
		}
		
		pMesh[A].data.Orientation = oldO;
		pMesh[A].data.Position    = oldP;
	}
		
	//Check if 2 or more vertices are intersecting
	//if so update choice of normals accordingly and choice of old held normal
	for(i=0;i<NoOfObjects;i++)
	for(j=0;j<NoOfObjects;j++)
	{
		if(i==j)
			continue;

		int k,m;
		int indexAiBj;
		int indexAjBi;
		int NoOfContactsFromAi;
		int NoOfContactsFromAj;
		int ArrayAi[4];
		int ArrayAj[4];
		int choice1,choice2;
		CVector vDifference;
		double  bestDistance,currDistance;

		//ObjectA=i && ObjectB=j occurs at
		indexAiBj = i*4*(NoOfObjects-1)+j*4;
		if(j>i)
			indexAiBj-=4;

		//ObjectA=j && ObjectB=i occurs at
		indexAjBi = j*4*(NoOfObjects-1)+i*4;
		if(i>j)
			indexAjBi-=4;

		NoOfContactsFromAi=0;
		NoOfContactsFromAj=0;
		for(k=0;k<4;k++)
		{
			if(Array[indexAiBj+k].bContactExists==true)
			{
				ArrayAi[NoOfContactsFromAi] = indexAiBj+k;
				NoOfContactsFromAi++;
			}

			if(Array[indexAjBi+k].bContactExists==true)
			{
				ArrayAj[NoOfContactsFromAj] = indexAjBi+k;
				NoOfContactsFromAj++;
			}
		}

		if(NoOfContactsFromAi+NoOfContactsFromAj<=1)
			continue;

		//***Find 2 vertices that represent the contact***

		//No Contacts from A = j && B = i
		//Choose two points from AiBj that are as far apart as possible
		if(NoOfContactsFromAj==0)
		{
			bestDistance = -1;
			for(k=0  ;k<NoOfContactsFromAi;k++)
			for(m=k+1;m<NoOfContactsFromAi;m++)
			{
				vDifference = Array[ArrayAi[k]].objectA_IntersectionPoint - Array[ArrayAi[m]].objectA_IntersectionPoint;
				currDistance = vDifference.Length();
				if(currDistance>bestDistance)
				{
					choice1 = ArrayAi[k];
					choice2 = ArrayAi[m];
					bestDistance = currDistance;
				}
			}
		}

		//No Contacts from A = i && B = j
		//Choose two points from AjBi that are as far apart as possible
		if(NoOfContactsFromAi==0)
		{
			bestDistance = -1;
			for(k=0  ;k<NoOfContactsFromAj;k++)
			for(m=k+1;m<NoOfContactsFromAj;m++)
			{
				vDifference = Array[ArrayAj[k]].objectA_IntersectionPoint - Array[ArrayAj[m]].objectA_IntersectionPoint;
				currDistance = vDifference.Length();
				if(currDistance>bestDistance)
				{
					choice1 = ArrayAj[k];
					choice2 = ArrayAj[m];
					bestDistance = currDistance;
				}
			}
		}

		//Contacts from A = i and j
		//Choose 1 point from AiBj and 1 from AjBi that are as far apart as possible
		if(NoOfContactsFromAi!=0 && NoOfContactsFromAj!=0)
		{
			bestDistance = -1;
			for(k=0;k<NoOfContactsFromAi;k++)
			for(m=0;m<NoOfContactsFromAj;m++)
			{
				vDifference = (pMesh[i].data.Orientation*Array[ArrayAi[k]].objectA_IntersectionPoint+pMesh[i].data.Position)
							- (pMesh[j].data.Orientation*Array[ArrayAj[m]].objectA_IntersectionPoint+pMesh[j].data.Position);
				currDistance = vDifference.Length();
				if(currDistance>bestDistance)
				{
					choice1 = ArrayAi[k];
					choice2 = ArrayAj[m];
					bestDistance = currDistance;
				}
			}
		}
		//We have our 2 vertices of contact

		//***Now we need to find an appropriate normal***
		CVector Point1;
		CVector Point2;
		CVector Normal1;
		CVector Normal2;

		//World Coordinates
		Point1 = pMesh[Array[choice1].idObjectA].data.Orientation*Array[choice1].objectA_IntersectionPoint+pMesh[Array[choice1].idObjectA].data.Position;
		Point2 = pMesh[Array[choice2].idObjectA].data.Orientation*Array[choice2].objectA_IntersectionPoint+pMesh[Array[choice2].idObjectA].data.Position;

		vDifference = Point1 - Point2;

		//Make sure the distance between the 2 points is not too small

		//Currently this is done by comparing it to the 2*cornerDistance
		//It may be better to replace it with something based on side lengths
		if(vDifference.Length()>2*cornerDistance)
		{
			Normal1 = vDifference.Cross(CVector(0,1,0));
			Normal1.e[01] = 0;
			Normal2 = Normal1;
			
			//ObjectB Coordinates
			Point1  = pMesh[Array[choice1].idObjectB].data.Orientation.ReturnInverse()*(Point1 - pMesh[Array[choice1].idObjectB].data.Position);
			Point2  = pMesh[Array[choice2].idObjectB].data.Orientation.ReturnInverse()*(Point2 - pMesh[Array[choice2].idObjectB].data.Position);
			
			Normal1 = pMesh[Array[choice1].idObjectB].data.Orientation.ReturnInverse()*Normal1;
			Normal2 = pMesh[Array[choice2].idObjectB].data.Orientation.ReturnInverse()*Normal2;
			
			//Check to see if this method will give a clear Normal
			//If so update choice of collision normal accordingly
			
			//5.83 = (tan(45*3/2))^2
			if(5.83*Normal1.e[0]*Normal1.e[0]<Normal1.e[2]*Normal1.e[2])
			{
				//Clearly Vertical
				if(Point1.e[2]>0)
					Array[choice1].objectB_OutwardNormal = CVector(0,0,1);
				else
					Array[choice1].objectB_OutwardNormal = CVector(0,0,-1);
			}
			
			if(Normal1.e[0]*Normal1.e[0]>5.83*Normal1.e[2]*Normal1.e[2])
			{
				//Clearly Horizontal
				if(Point1.e[0]>0)
					Array[choice1].objectB_OutwardNormal = CVector(1,0,0);
				else
					Array[choice1].objectB_OutwardNormal = CVector(-1,0,0);
			}
			
			if(5.83*Normal2.e[0]*Normal2.e[0]<Normal2.e[2]*Normal2.e[2])
			{
				//Clearly Vertical
				if(Point2.e[2]>0)
					Array[choice2].objectB_OutwardNormal = CVector(0,0,1);
				else
					Array[choice2].objectB_OutwardNormal = CVector(0,0,-1);
			}
			
			if(Normal2.e[0]*Normal2.e[0]>5.83*Normal2.e[2]*Normal2.e[2])
			{
				//Clearly Horizontal
				if(Point2.e[0]>0)
					Array[choice2].objectB_OutwardNormal = CVector(1,0,0);
				else
					Array[choice2].objectB_OutwardNormal = CVector(-1,0,0);
			}
		}

		//If from the same object they must have the same normal
		if(Array[choice1].idObjectA == Array[choice2].idObjectA)
			Array[choice2].objectB_OutwardNormal = Array[choice1].objectB_OutwardNormal;

		//Update the contactExists and both normal info
		for(k=0;k<4;k++)
		{
			if(Array[indexAiBj+k].bContactExists==true)
			{
				Array[indexAiBj+k].bContactExists = false;
				if(Array[indexAiBj+k].idObjectA==Array[choice1].idObjectA)
				{
					Array[indexAiBj+k].objectB_OutwardNormal = Array[choice1].objectB_OutwardNormal;
					Array[indexAiBj+k].Extra.old_Normal      = Array[choice1].objectB_OutwardNormal;
				}
				else
				{
					Array[indexAiBj+k].objectB_OutwardNormal = Array[choice2].objectB_OutwardNormal;
					Array[indexAiBj+k].Extra.old_Normal      = Array[choice2].objectB_OutwardNormal;
				}
			}

			if(Array[indexAjBi+k].bContactExists==true)
			{
				Array[indexAjBi+k].bContactExists = false;
				if(Array[indexAjBi+k].idObjectA==Array[choice1].idObjectA)
				{
					Array[indexAjBi+k].objectB_OutwardNormal = Array[choice1].objectB_OutwardNormal;
					Array[indexAjBi+k].Extra.old_Normal      = Array[choice1].objectB_OutwardNormal;
				}
				else
				{
					Array[indexAjBi+k].objectB_OutwardNormal = Array[choice2].objectB_OutwardNormal;
					Array[indexAjBi+k].Extra.old_Normal      = Array[choice2].objectB_OutwardNormal;
				}
			}
		}

		Array[choice1].bContactExists = true;
		Array[choice2].bContactExists = true;

	}//End of check for 2 or more points involved in an intersection

	return;
}



//Our main collision response routine
void UpdateMotionData(C3dObject* pMesh, int NoOfObjects)
{
	static bool firstTime = true;

	if(firstTime==true)
	{
		InitializeCContactArray(pMesh,NoOfObjects,contactArray);
		TestCuboidIntersection(pMesh,NoOfObjects,contactArray);
		firstTime = false;
	}

	int         i,j;
	double      t = Global_T;
	CVector     g;
	CVectorList vlStart, vlEnd;

	if(!vlStart.setDimension(10*NoOfObjects))
	{
		printf("*** Ran Out Of Memory ***\n");
		return;
	}

	if(!vlEnd.setDimension(10*NoOfObjects))
	{
		printf("*** Ran Out Of Memory ***\n");
		return;
	}

	g.e[0] = 0;
	g.e[1] = 0;
	g.e[2] = -9.8*Global_UnitsPerMetre;


	//Store current state as old state
	for(i=0;i<NoOfObjects;i++)
	{
		//Get rid of all y axis motion
		pMesh[i].data.new_Position.e[1] = 0;
		pMesh[i].data.new_Velocity.e[1] = 0;

		pMesh[i].data.old_AngularMomentum = pMesh[i].data.new_AngularMomentum;
		pMesh[i].data.old_Orientation     = pMesh[i].data.new_Orientation;
		pMesh[i].data.old_Position        = pMesh[i].data.new_Position;
		pMesh[i].data.old_Velocity        = pMesh[i].data.new_Velocity;

		pMesh[i].data.Position=pMesh[i].data.old_Position;
		pMesh[i].data.Orientation.Rotation(0,1,0,pMesh[i].data.old_Orientation);
	}

	
	//Resolve All Impulses
	//A contributes a vertex, B the normal
	int     Ai,   Bi;
	double  Am,   Bm;
	CVector Av,   Bv;
	double  AL,   BL;
	double  AInv, BInv;
	CVector Aw,   Bw;
	CVector Ar,   Br;
	CVector Arxn, Brxn;
	
	CVector r,n;


	TestCuboidIntersection(pMesh,NoOfObjects,contactArray);

	i=0;
	while(i<NoOfObjects*(NoOfObjects-1))
	{
		//***Resolve Any collisions between Ai and Bi***

		Ai   = contactArray[i*4].idObjectA;
		Bi   = contactArray[i*4].idObjectB;

		//Update m,Iinv
		Am   = pMesh[Ai].data.Mass;
		AInv = 1/pMesh[Ai].data.Inertia;
		
		Bm   = pMesh[Bi].data.Mass;
		BInv = 1/pMesh[Bi].data.Inertia;

		//Check if testing two Infinite Masses colliding
		if(Am<=0 && Bm<=0)
		{
			i++;
			continue;
		}

		//If we had to apply an impulse we'll have to restart the entire loop from i=0
		//but before we do that lets resolve all impulses between Ai and Bi
		bool bNeedToRestart = false;
		int  iImpulsesApplied;

		do
		{
			iImpulsesApplied = 0;
			
			for(j=0;j<4;j++)
			{
				contactArray[i*4+j].Extra.bSmallVelocityContact = false; //Used Later

				if(contactArray[i*4+j].bContactExists == false)
					continue;

				//***Calculate Separation Velocity***
				
				//Update v,L,w
				Av = pMesh[Ai].data.new_Velocity;
				AL = pMesh[Ai].data.new_AngularMomentum;
				Aw = AInv*AL*CVector(0,1,0);
				
				Bv = pMesh[Bi].data.new_Velocity;
				BL = pMesh[Bi].data.new_AngularMomentum;
				Bw = BInv*BL*CVector(0,1,0);
				
				//Update r,n
				r       = pMesh[Ai].data.Orientation*contactArray[i*4+j].objectA_IntersectionPoint+pMesh[Ai].data.Position;
				n       = pMesh[Bi].data.Orientation*contactArray[i*4+j].objectB_OutwardNormal;
				n.e[1]  = 0;
				n.Normalize();

				Ar      = r-pMesh[Ai].data.new_Position;
				Ar.e[1] = 0;

				Br      = r-pMesh[Bi].data.new_Position;
				Br.e[1] = 0;

				
				//Calculate Velocities
				CVector ObjVelA;
				CVector ObjVelB;
				
				ObjVelA = Av+Aw.Cross(Ar);
				ObjVelB = Bv+Bw.Cross(Br);

				double SepVelocity = (ObjVelA-ObjVelB).Dot(n);

				if(Global_VelocityZero<SepVelocity)
					contactArray[i*4+j].Extra.bSmallVelocityContact = false; //Used Later
				else
					contactArray[i*4+j].Extra.bSmallVelocityContact = true;  //Used Later

				if(SepVelocity<-0.0000001)
				{
					//***Apply An Impulse***
					double J;
					double e;
					
					e   = pMesh[Ai].data.CoefficientOfRestitution
						* pMesh[Bi].data.CoefficientOfRestitution;
					
					Arxn = Ar.Cross(n);
					Brxn = Br.Cross(n);
					
					double Am_reciprocal;
					double Bm_reciprocal;

					if(Am<=0)
						Am_reciprocal=0;
					else
						Am_reciprocal=1/Am;
					
					if(Bm<=0)
						Bm_reciprocal=0;
					else
						Bm_reciprocal=1/Bm;
					

					J = -(1+e) * (Av.Dot(n)+Arxn.Dot(Aw)-Bv.Dot(n)-Brxn.Dot(Bw))
						/ (Am_reciprocal*(1+Arxn.Dot(AInv*Arxn))
						+  Bm_reciprocal*(1+Brxn.Dot(BInv*Brxn)));
					
					
					pMesh[Ai].data.new_Velocity       +=Am_reciprocal*J*n;
					pMesh[Ai].data.new_AngularMomentum+=Am_reciprocal*J*Arxn.e[1];
					pMesh[Bi].data.new_Velocity       -=Bm_reciprocal*J*n;
					pMesh[Bi].data.new_AngularMomentum-=Bm_reciprocal*J*Brxn.e[1];
										
					iImpulsesApplied++;
					bNeedToRestart = true;
				}
				
			}//End of For Loop

		}while(iImpulsesApplied>0);

		if(bNeedToRestart)
			i=0;
		else
			i++;
	}//End of Resolving All Impulses

	//Allow Free Motion
	CDiffEqMotionData* MD = mdArray;

	for(i=0;i<NoOfObjects;i++)
	{
		MD[i].Mass            = pMesh[i].data.Mass;
		MD[i].Inertia         = pMesh[i].data.Inertia;
		MD[i].Orientation     = pMesh[i].data.new_Orientation;
		MD[i].Position        = pMesh[i].data.new_Position;
		MD[i].AngularMomentum = pMesh[i].data.new_AngularMomentum;
		MD[i].Velocity        = pMesh[i].data.new_Velocity;
	}

	MotionDataToArray(vlStart.array,MD,NoOfObjects);
	
	//Numerically integrate the state forward in time
	//The equations of motion are given by the function qDash()
	ode(&vlStart,&vlEnd,0,t,1,qDash);

	ArrayToMotionData(vlEnd.array,MD,NoOfObjects);

	//Update the state of the object
	for(i=0;i<NoOfObjects;i++)
	{
		if(pMesh[i].data.Mass>0)
		{
			pMesh[i].data.new_Orientation     = MD[i].Orientation;
			pMesh[i].data.new_Position        = MD[i].Position;
			pMesh[i].data.new_AngularMomentum = MD[i].AngularMomentum;
			pMesh[i].data.new_Velocity        = MD[i].Velocity;
		}
			
		pMesh[i].data.new_Position.e[1] = 0;
		pMesh[i].data.new_Velocity.e[1] = 0;

		pMesh[i].data.Position = pMesh[i].data.new_Position;
		pMesh[i].data.Orientation.Rotation(0,1,0,pMesh[i].data.new_Orientation);	
	}

	return;
}

