//*************************************************************************************
//This program simulates an object colliding with the floor, without friction.
//Note that this program does not handle sliding off the edge of the checkered floor.
//
//The object which we are simulating and its properties are set in the init() function.
//Change the string in the Load3dsFile() function for Object[1] to change the boundary
//of the object. Change the 'data' member of Object[1] to change its simulation
//properties. The gravity strength and time step variables of the simulation can be
//changed in the file MotionData.cpp.
//
//To change the camera angle left click and drag the mouse within the simulation
//window.
//
//Videos can be created but only when running in a Windows operating system which has
//DirectX 9.0b or higher installed. Whilst creating video do not resize the simulation
//window as this will corrupt the image. On start up the program will ask how you wish
//the video to be compressed. When you wish to stop recording merely close the opengl
//simulation window (not the console window). The resulting video will appear in a
//folder placed in the same directory as the program.
//*************************************************************************************


//This program uses OpenGl to render the simulation.
//Include the following libraries when building in windows:
//opengl32.lib glu32.lib glut32.lib

//If _WIN_ is defined then the application will record the simulation into a video
//This option is only available when compiling with directX in Windows
#define _WIN_
#undef  _WIN_//Comment this line out to record videos ofthe simulation

#ifdef _WIN_
#include "..//VideoCapture.h"
#include <windows.h>
#include "..//resource.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <GL/glut.h>

#include "3ds_loader.h"
#include "Quaternion.h"

//The width and height of the window
//You should not resize the window while creating a video
int screen_width=400;
int screen_height=400;
int sw,sh;

//Video variables
int    MsPerFrame=20;

//Mouse Values
int    initX;
int    initY;

//Camera Angles
double initZCamRot;
double initXCamRot;
double zCamRot;
double xCamRot;

//Lighting Position
double lx=0,ly=-200,lz=200;

//Camera Position
double cx=0,cy=-400,cz=50;
 
//Flags for rendering
int filling = 1; //Press 'R' to change: 0=Solid   1=Wireframe
int flat    = 1; //Press 'F' to change: 0=Gouraud 1=Flat

//Models
int       NoOfObjects = 2;
C3dObject Object[2];
C3dObject TileWhite;
C3dObject TileBlack;
bool      bLoadedModels;//If the models couldn't be loaded this is set to false


#ifdef _WIN_
//Video Capture

VideoProperties vp;
ResourceValues  rv;
CVideoCapture   cVC;

BYTE* pData = NULL;
#endif


//This is where we initialize our scene and simulation
void init(void)
{
	int i;

    glClearColor(0.0, 0.0, 0.0, 0.0); //Background Color is Black
    glShadeModel(GL_SMOOTH); //Type of shading for the polygons

    // Viewport transformation
    glViewport(0,0,screen_width,screen_height);  

    // Projection transformation
    glMatrixMode(GL_PROJECTION); 
    glLoadIdentity();
    gluPerspective(45.0f,(GLfloat)screen_width/(GLfloat)screen_height,10.0f,10000.0f);
   
    glEnable(GL_DEPTH_TEST); //Enable Z-Buffer
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL); //Render Polygons as filled    

	//Lighting
	float position [] = {lx,ly,lz,1};
	float diffuse  [] = {0.8,0.8,0.8,1};
	float ambient  [] = {0.2,0.2,0.2,1}; 
	float specular [] = {0,0,0,1};
	float black    [] = {0,0,0,0};
	float other    [] = {1.0,0,0,0};
	
	glEnable        (GL_LIGHTING);
	glLightModeli   (GL_LIGHT_MODEL_TWO_SIDE, 0);

	glEnable        (GL_COLOR_MATERIAL);
	glColorMaterial (GL_FRONT,  GL_DIFFUSE);
	glMaterialfv    (GL_FRONT,  GL_AMBIENT,black);
	glMaterialfv    (GL_FRONT,  GL_SPECULAR,black);

	glLightfv       (GL_LIGHT0, GL_POSITION, position);
	glLightfv       (GL_LIGHT0, GL_DIFFUSE,  diffuse);
	glLightfv       (GL_LIGHT0, GL_AMBIENT,  ambient);
	glLightfv       (GL_LIGHT0, GL_SPECULAR, specular);
	glEnable        (GL_LIGHT0);

	//Initialize Camera Angles
	zCamRot = 0;
	xCamRot = 0;


	//Load Model data
	printf("Loading Models...\n");



	//The model name
	if(!Load3dsFile(&(Object[1]),"Models/cone.3ds"))//Can change this to any model in the Models folder
	{
		printf("Error! Couldn't load 3ds object for simulation.\n");
		bLoadedModels=false;
		return;
	}
	printf("\n");
	printf("Name     : %s\n",Object[1].name);
	printf("Vertices : %i\n",Object[1].iNumOfVerts);
	printf("Faces    : %i\n",Object[1].iNumOfFaces);
	for(i=0;i<Object[1].iNumOfVerts;i++)
	{
		Object[1].vertList[i].r = 1;
		Object[1].vertList[i].g = 0;
		Object[1].vertList[i].b = 0;
	}

	//Set the object's properties

	//Automatically find the object's inertia tensor assuming it was uniform solid and move the 
	//	object so its centre of mass is at the origin in its model space.
	//Change this to CalculateHollowInertiaAndCOM() for hollow objects
	//You can also manually set the InertiaTensor 
	//	it expects the inertia tensor for the object as if its total mass was 1
	CalculateSolidInertiaAndCOM(&(Object[1]));
	
	//Set the initial state of the object
	Object[1].data.new_AngularMomentum=Object[1].data.InertiaTensor*CVector(0.3,0.1,1.0);
	Object[1].data.new_Orientation = CQuaternion(0,1,0,0);
	Object[1].data.new_Velocity = CVector(0,0,0);
	Object[1].data.new_Position = CVector(0,0,100);

	//Set the Coefficient of Restitution
	Object[1].data.CoefficientOfRestitution = 0.2;

	//The Total Mass of the object 
	Object[1].data.Mass = 1;

	



	//Load the object which will act as the floor
	if(!Load3dsFile(&(Object[0]),"floor.3ds"))
	{
		printf("Error! Couldn't load floor.3ds, make sure its in the same directory as this program.\n");
		bLoadedModels=false;
		return;
	}
	printf("\n");
	printf("Name     : %s\n",Object[0].name);
	printf("Vertices : %i\n",Object[0].iNumOfVerts);
	printf("Faces    : %i\n",Object[0].iNumOfFaces);

	//Set the floor's properties (stationary and immovable)
	CalculateSolidInertiaAndCOM(&(Object[0]));
	Object[0].data.new_AngularMomentum=Object[0].data.InertiaTensor*CVector(0,0,0);//Stationary
	Object[0].data.new_Orientation = CQuaternion(1,0,0,0);
	Object[0].data.new_Velocity = CVector(0,0,0);//Stationary
	Object[0].data.new_Position = CVector(0,0,-250);
	Object[0].data.CoefficientOfRestitution = 1;
	Object[0].data.Mass = -1;//Immovable

	//Load Black and White Tiles that will be used to display a checkered floor 
	if(!Load3dsFile(&TileWhite,"tile.3ds"))
	{
		printf("Error! Couldn't load tile.3ds, make sure its in the same directory as this program.\n");
		bLoadedModels=false;
		return;
	}
	printf("\n");
	printf("Name     : %s\n",TileWhite.name);
	printf("Vertices : %i\n",TileWhite.iNumOfVerts);
	printf("Faces    : %i\n",TileWhite.iNumOfFaces);
	for(i=0;i<TileWhite.iNumOfVerts;i++)
	{
		TileWhite.vertList[i].r = 1.0;
		TileWhite.vertList[i].g = 1.0;
		TileWhite.vertList[i].b = 1.0;
	}
	
	if(!Load3dsFile(&TileBlack,"tile.3ds"))
	{
		printf("Error! Couldn't load tile.3ds, make sure its in the same directory as this program.\n");
		bLoadedModels=false;
		return;
	}
	printf("\n");
	printf("Name     : %s\n",TileBlack.name);
	printf("Vertices : %i\n",TileBlack.iNumOfVerts);
	printf("Faces    : %i\n",TileBlack.iNumOfFaces);
	for(i=0;i<TileBlack.iNumOfVerts;i++)
	{
		TileBlack.vertList[i].r = 0.4;
		TileBlack.vertList[i].g = 0.4;
		TileBlack.vertList[i].b = 0.4;
	}

	printf("\n");

	//Allocate memory needed by our update simulation routine
	if(!InitGlobalVariables(Object[1].iNumOfVerts))
	{
		printf("Error! Couldn't allocate memory for contact resolution.\n");
		bLoadedModels=false;
		return;
	}

	bLoadedModels=true;
	return;
}



//OpenGL function that handles window resizing
void resize (int width, int height)
{
    screen_width=width;
    screen_height=height;

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glViewport(0,0,screen_width,screen_height);

    glMatrixMode(GL_PROJECTION); 
    glLoadIdentity();
    gluPerspective(45.0f,(GLfloat)screen_width/(GLfloat)screen_height,10.0f,10000.0f);

    glutPostRedisplay(); //Redraw the window 
}



//OpenGl function to handle keyboard input
void keyboard (unsigned char key, int x, int y)
{
    switch (key)
    {
		//Change between Wire Frame and Solid
        case 'r': case 'R':
            if (filling==0)
            {
                glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
                filling=1;
            }   
            else 
            {
                glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
                filling=0;
            }
        break;

		//Change between Flat shading and Gouraud
		case 'f': case 'F':
			flat=1-flat;
			break;
	}
}



//OpenGl function to handle mouse clicks
void mouseClick(int button, int state, int x, int y)
{
	//button  = GLUT_LEFT_BUTTON
	//          GLUT_MIDDLE_BUTTON
	//          GLUT_RIGHT_BUTTON
	//
	//state   = GLUT_UP
	//          GLUT_DOWN
	//
	//x, y    = Window relative coordinates
	
	if(button!=GLUT_LEFT_BUTTON)
		return;

	if(state==GLUT_DOWN)
	{
		initX       = x;
		initY       = y;
		initZCamRot = zCamRot;
		initXCamRot = xCamRot;

	}

	return;
}



//OpenGl function to handle mouse movement
void mouseMove(int x, int y)
{
	xCamRot = initXCamRot + (y-initY)*0.2;
	zCamRot = initZCamRot + (x-initX)*0.2;
	return;
}



//Draws shadow on tiled floor by simple projecion
void DrawShadow(C3dObject* pMesh)
{
	int i,j;
	CVertex v;
	CVector x,xNew[3]; //Vertex Positions
	CVector L;         //Light Position
	CVector n;         //Normal to Floor
	double nDotL;
	double t,denom;
	bool bDraw;

	L     = CVector(lx,ly,lz);
	n     = CVector(0,0,1);
	nDotL = n.Dot(L);//Assume +ve ie. light above plane

	glDisable(GL_LIGHTING);

	glBegin(GL_TRIANGLES);
	{
		for(i=0;i<pMesh->iNumOfFaces;i++)
		{
			bDraw = true;
			for(j=0;j<3;j++)
			{	
				v = pMesh->Transform(pMesh->faceList[i].index[j]);
				x.e[0] = v.x;
				x.e[1] = v.y;
				x.e[2] = v.z;
				
				denom = n.Dot(x) - nDotL;
				if(denom>-0.0001)
				{
					bDraw = false;
					break;
				}
				
				t=-nDotL/denom;
				xNew[j]=L*(1-t)+x*t;
			}

			if(bDraw==false)
				continue;

			for(j=0;j<3;j++)
			{
				glColor3f(0,0,0);
				glVertex3f(xNew[j].e[0],xNew[j].e[1],1);
				//z = 1 to avoid z-fighting
			}
		}
	}
    glEnd();

	glEnable(GL_LIGHTING);

	return;
}



//Draw our mesh object
void DrawObject(C3dObject* pMesh, int ForceFlat=0)
{
	int i,j;
	CVertex v;
	CVector n;

	glBegin(GL_TRIANGLES);
	{
		
		for(i=0;i<pMesh->iNumOfFaces;i++)
		{
			n = pMesh->data.Orientation * pMesh->faceList[i].Normal;

			for(j=0;j<3;j++)
			{	
				v = pMesh->Transform(pMesh->faceList[i].index[j]);

				if(flat==1 || ForceFlat==1)
					glNormal3f(n.e[0],n.e[1],n.e[2]);
				else
					glNormal3f(v.Normal.e[0],v.Normal.e[1],v.Normal.e[2]);
					
				glColor3f(v.r,v.g,v.b);
				glVertex3f(v.x,v.y,v.z);
			}
		}

	}
    glEnd();

	return;
}



//Apply before Model Transform
void ApplyCameraTransform()
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
	
	glRotatef(-90,1,0,0);
	glTranslatef(-cx,-cy,-cz);
	glRotatef(xCamRot,1,0,0);
	glRotatef(zCamRot,0,0,1);
	return;
}



//The main rendering routine
//This is called every frame
void display(void)
{
    int   i,j;
	int   noTilesX   = 10;
	int   noTilesY   = 10;
	float dimension  = 50;
	float ambient[]  = { 0, 0, 0, 1};
	float position[] = {lx,ly,lz, 1}; //Light Position

	//Update the simulation
	//This function calculates the objects motion
	UpdateMotionData(Object,NoOfObjects);

	//Set up the scene
	ApplyCameraTransform();
	glLightfv(GL_LIGHT0, GL_POSITION, position);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	ApplyCameraTransform();

	//Draw our object and its shadow into the Back Buffer
	ambient[0]=1;
	ambient[1]=0;
	ambient[2]=0;
	glMaterialfv(GL_FRONT,GL_AMBIENT,ambient);
	DrawObject(&(Object[1]));
	DrawShadow(&(Object[1]));


	//Draw checkered floor [dimension of tiles 50 x 50 x -5]
	for(i=0;i<noTilesX;i++)
		for(j=0;j<noTilesY;j++)
		{
			ApplyCameraTransform();

			glTranslatef(i*dimension-(noTilesX-1)*dimension/2.0,j*dimension-(noTilesY-1)*dimension/2.0,0);
			
			if((i+j)%2==0)
			{
				ambient[0]=1;
				ambient[1]=1;
				ambient[2]=1;
				glMaterialfv(GL_FRONT,GL_AMBIENT,ambient);
				DrawObject(&TileWhite,1);
			}
			else
			{		
				ambient[0]=0;
				ambient[1]=0;
				ambient[2]=0;
				glMaterialfv(GL_FRONT,GL_AMBIENT,ambient);
				DrawObject(&TileBlack,1);
			}
		}

    glFlush(); //Force OpenGl commands to be executed
    glutSwapBuffers(); //Display Back Buffer

#ifdef _WIN_
	//Video Capture

	//Take a screen grab
	glFinish();

	glPixelStorei(GL_PACK_ALIGNMENT, 4);
	glPixelStorei(GL_PACK_ROW_LENGTH, 0);
	glPixelStorei(GL_PACK_SKIP_ROWS, 0);
	glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
	glReadPixels(0, 0, sw,sh, GL_BGR_EXT, GL_UNSIGNED_BYTE, pData);

	//Add the image to our video
	while(!cVC.AddImage(pData,MsPerFrame))
		Sleep(1);
#endif

}



//The main loop
int main(int argc, char **argv)
{
	sw=screen_width;
	sh=screen_height;
 
#ifdef _WIN_
	//Video Capture

	//The size of the Image Buffer
	int size   = screen_height*(((screen_width*24)/8-1)/4+1)*4;

	//Set Video Properties
	vp.format             = RGB_24;
	vp.width              = screen_width;
	vp.height             = screen_height;
	
	vp.bCreateBuffer      = true;
    vp.bUseTheseSettings  = false;

	vp.bWriteOverIfExists = true;
	vp.name               = "videos/Output.avi";

	//Set the Resource Values
	rv.Set(
		IDD_COMPRESS_DIALOG,
		IDC_COMBO,
		IDC_CONFIGURE,
		IDC_ABOUT,
		IDOK,
		IDC_SCROLLBAR,
		IDC_QUALITY_TEXT,
		IDC_VALUE
		);

	//Intialize CVideoCapture
	cVC.Create();

	cVC.SetResourceValues(&rv);
	cVC.InitArray();
	cVC.DisplayCompressorDialog(NULL,&vp);
	
	cVC.Init();

	printf("\n-----------------------\n\n");

	//Create Image Buffer
	pData=NULL;
	pData=new BYTE[size];
	if(pData==NULL)
	{
		printf("Error! Ran out of Memory\n");
		cVC.Cleanup();
		cVC.Destroy();
		return 1;
	}

	//Create Filter Graph
    if(SUCCEEDED(cVC.CreateFilterGraph(&vp)))
	{
		printf("Filter Graph Created\n");
		if(size==cVC.GetBufferSize())
			printf("Size Of Buffer Consistent\n\n");
		else
		{
			printf("Error! Buffer Size Not Expected\n");
			cVC.Cleanup();
			cVC.Destroy();
			delete[] pData;
			pData=NULL;
			return 1;
		}
	}
	else
	{
		printf("Error! Filter Graph Failed\n");
		cVC.Cleanup();
		cVC.Destroy();
		delete[] pData;
		pData=NULL;
		return 1;
	}
#endif

    //We use the GLUT utility to handle the window
    glutInit(&argc, argv);    
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(screen_width,screen_height);
    glutInitWindowPosition(0,0);
    glutCreateWindow("Floor Collision");
    
    glutDisplayFunc  (display);
    glutIdleFunc     (display);
    glutReshapeFunc  (resize);
    glutKeyboardFunc (keyboard);
	glutMouseFunc    (mouseClick);
	glutMotionFunc   (mouseMove);

	init();
	if(bLoadedModels==true)
		glutMainLoop();//main window loop

	//Clean up
	ReleaseGlobalVariables();

#ifdef _WIN_
	cVC.Cleanup();
	cVC.Destroy();
	delete[] pData;
	pData=NULL;
#endif

    return 0;    
}

