#include "MotionData.h"
#include "3d_object.h"
#include "ContactInfo.h"
#include "LinearProgramSolver.h"
#include <stdio.h>


double     Global_T             = 0.1;  //Time Step
double     Global_UnitsPerMetre = 0.2;  //Use this to scale the effects of gravity
double     Global_VelocityZero  = 0.2;  //Threshold velocity used in deciding which contacts are resting contacts
C3dObject* Global_pMesh;
int        sizeOfArray          = 0;    //We make the arrays as big as the number of vertices the object has
CContact*  contactArray         = NULL; //One possible contact per vertex
int*       restingContactPoints = NULL; //Should be same size as contactArray[]


//Constructor of CMotionData
//Just sets some sensible default values for the properties
CMotionData::CMotionData()
{
	Orientation.Identity();
	Position = CVector(0,0,0);

	InertiaTensor.Identity();
	InverseOfTheInertiaTensor.Identity();

	new_Orientation     = old_Orientation;
	new_Position        = old_Position;
	new_AngularMomentum = old_AngularMomentum;
	new_Velocity        = old_Velocity;
}



int InitGlobalVariables(int numOfVertices)
{
	sizeOfArray          = 0;
	contactArray         = NULL;
	restingContactPoints = NULL;

	contactArray = new CContact[numOfVertices];
	if(contactArray==NULL)
		return 0;

	restingContactPoints = new int[numOfVertices];
	if(restingContactPoints==NULL)
	{
		delete[] contactArray;
		contactArray=NULL;
		return 0;
	}

	sizeOfArray = numOfVertices;
	return 1;
}

void ReleaseGlobalVariables()
{
	if(contactArray!=NULL)
		delete[] contactArray;

	if(restingContactPoints!=NULL)
		delete[] restingContactPoints;

	sizeOfArray          = 0;
	contactArray         = NULL;
	restingContactPoints = NULL;

	return;
}









//Used to convert the output from integrating motion back into a state of the simulation
void ArrayToState(double* array, CQuaternion& Orientation, CVector& L, CVector& Position, CVector& Velocity, CMatrix& M)
{
	int i;

	for(i=0;i<4;i++)
		Orientation.e[i] = array[i];

	for(i=0;i<3;i++)
		L.e[i]           = array[i+4];

	for(i=0;i<3;i++)
		Position.e[i]    = array[i+7];

	for(i=0;i<3;i++)
		Velocity.e[i]    = array[i+10];

	for(i=0;i<9;i++)
		M.M[i/3][i%3]    = array[i+13];

	return;
}

//Used to convert the state of an object into an array which can then be integrated
void StateToArray(double* array, CQuaternion& Orientation, CVector& L, CVector& Position, CVector& Velocity, CMatrix& M)
{
	int i;

	for(i=0;i<4;i++)
		array[i]    = Orientation.e[i];

	for(i=0;i<3;i++)
		array[i+4]  = L.e[i];

	for(i=0;i<3;i++)
		array[i+7]  = Position.e[i];

	for(i=0;i<3;i++)
		array[i+10] = Velocity.e[i];

	for(i=0;i<9;i++)
		array[i+13] = M.M[i/3][i%3];

	return;
}



//This function gives the equations of motion that needed to be integrated
void qDash(CVectorList* x_dot, CVectorList* x, double t)
{
	int i,j;
	CMatrix R;
	CQuaternion q,qdot;
	CVector w;
	CMatrix Iinverse,Iinv;
	CVector L;
	CVector P;
	CVector V;

	x_dot->Fill(0);


	ArrayToState(x->array,q,L,P,V,Iinverse);

	QuaternionToMatrix(q,R);
	Iinv = R*Iinverse*R.ReturnTranspose();
	w    = Iinv*L;
	qdot = 0.5*w*q;

	//Array holds a list of resting contact points (renamed to make the code more readable)
	//F holds the number of resting contact points
	int* array = restingContactPoints;
	int  F;

	//Fill in contact point list, array[]
	F=0;
	for(i=0;i<sizeOfArray;i++)
		if(contactArray[i].bContactExists)
		{
			array[F]=i;
			F++;
		}


	CMatrixList mN,mRXN,mInv,mResult1,mResult2,mG,mF,mT,mLP,mWXL,mNWWR;
	CVector     wxL;
	
	if(!mF.setDimensions(3,1))
	{
		printf("*** Ran Out Of Memory ***\n");
		return;
	}
	if(!mT.setDimensions(3,1))
	{
		printf("*** Ran Out Of Memory ***\n");
		return;
	}

	//Find the torque mT, and the forces mF needed to update the state 
	if(F==0)
	{
		//Freefall motion
		mF.SetValue(1,1,0);
		mF.SetValue(2,1,0);
		mF.SetValue(3,1,-9.8*Global_UnitsPerMetre);

		mT.SetValue(1,1,0);
		mT.SetValue(2,1,0);
		mT.SetValue(3,1,0);
	}
	else
	{
		//Use constraint-based methods to find the magnitude of forces
		if(!mN.setDimensions(3,F))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		for(i=1;i<=F;i++)
		for(j=1;j<=3;j++)
		{
			mN.SetValue(j,i,contactArray[array[i-1]].Normal.e[j-1]);
		}

		if(!mNWWR.setDimensions(F,1))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		if(!mRXN.setDimensions(3,F))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		for(i=1;i<=F;i++)
		{
			CVector r,n,rxn;
			r.e[0] = Global_pMesh->vertList[array[i-1]].x;
			r.e[1] = Global_pMesh->vertList[array[i-1]].y;
			r.e[2] = Global_pMesh->vertList[array[i-1]].z;
			r = R*r;
			n = contactArray[array[i-1]].Normal;
			rxn = r.Cross(n);

			mNWWR.SetValue(i,1,n.Dot(w.Cross(w.Cross(r))));

			for(j=1;j<=3;j++)
			{
				mRXN.SetValue(j,i,rxn.e[j-1]);
			}
		}

		if(!mInv.setDimensions(3,3))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		for(i=1;i<=3;i++)
		for(j=1;j<=3;j++)
		{
			mInv.SetValue(i,j,Iinv.M[i-1][j-1]);
		}

		if(!mG.setDimensions(3,1))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}
		
		if(!mWXL.setDimensions(3,1))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		mG.  SetValue(1,1,0);
		mG.  SetValue(2,1,0);
		mG.  SetValue(3,1,-9.8*Global_UnitsPerMetre);

		wxL = w.Cross(L);

		mWXL.SetValue(1,1,wxL.e[0]);
		mWXL.SetValue(2,1,wxL.e[1]);
		mWXL.SetValue(3,1,wxL.e[2]);


		mResult1 = mN.ReturnTranspose()*mN+mRXN.ReturnTranspose()*mInv*mRXN;
		mResult2 = -1*mN.ReturnTranspose()*mG+mRXN.ReturnTranspose()*mInv*mWXL-mNWWR;
		
		//Check if mResult1,2 is 0x0
		mResult1.getDimensions(&i,&j);
		if(i==0 || j==0)
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}
		
		mResult2.getDimensions(&i,&j);
		if(i==0 || j==0)
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		//Formulate the Linear Program
		//Find Feasible F satisfying:
		//mResult1*vF>=mResult2
		//Identity*vF>=0
		if(!mLP.setDimensions(F+1+F,F+2))
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}

		mLP.Fill(0);
		for(i=1;i<=F;i++)
		for(j=1;j<=F;j++)
			mLP.SetValue(i+1,j,mResult1.GetValue(i,j));

		for(i=1;i<=F;i++)
		{
			mLP.SetValue(i+1,F+1,-1);
			mLP.SetValue(i+1,F+2,mResult2.GetValue(i,1));

			mLP.SetValue(F+1+i,i,1);
			mLP.SetValue(F+1+i,F+1,-1);
		}

		CMatrixList vF;
		//Try all combinations of elements of vF forced to be zero
		//(We are trying to find a feasible set of vanishing contact points)
		int NoOfChoices=1;
		for(i=1;i<=F;i++)
			NoOfChoices*=2;

		for(i=0;i<=NoOfChoices-1;i++)
		{
			//Set elements in vF that will be zero
			int temp = i;
			for(j=1;j<=F;j++)
			{
				if(temp%2==0)
				{
					mLP.SetValue(F+1+j,F+1,-1);
					mLP.SetValue(j+1,F+1,0);
				}
				else
				{
					mLP.SetValue(F+1+j,F+1,0);
					mLP.SetValue(j+1,F+1,-1);
				}

				temp/=2;
			}

			//Find vF
			int returnResult;
			returnResult = SimplexAlgorithm(mLP,vF);

			//success
			if(returnResult==1)
				break;

			if(returnResult==-1)
			{
				printf("*** Ran Out Of Memory ***\n");
				return;
			}

			switch(returnResult)
			{
			case -2:
				//Should occur rarely
				//Set of vanishing contact points was unfeasible
				if(i==NoOfChoices-1)
					//Should never occur
					printf("No set of Vanishing Contact Points give a Feasible Solution\n");
				break;

			case -3:
				//Should never occur
				printf("Solution Is Unbounded\n");
				break;

			case -4:
				//Should never occur
				printf("Formulation Of Linear Program Is Too Small\n");
				break;

			default:
				//Should never occur
				printf("Error Solver Returned %i\n",returnResult);
			}
		}

		//Found vF
		mF = mG+(mN*vF);
		mT = mRXN*vF;
		
		//Check if mF, mT is 0x0
		mF.getDimensions(&i,&j);
		if(i==0 || j==0)
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}
		
		mT.getDimensions(&i,&j);
		if(i==0 || j==0)
		{
			printf("*** Ran Out Of Memory ***\n");
			return;
		}
	}


	x_dot->Fill(0);
	
	//update q
	x_dot->array[0]  = qdot.e[0];
	x_dot->array[1]  = qdot.e[1];
	x_dot->array[2]  = qdot.e[2];
	x_dot->array[3]  = qdot.e[3];

	//update Angular Momentum
	x_dot->array[4]  = mT.GetValue(1,1);
	x_dot->array[5]  = mT.GetValue(2,1);
	x_dot->array[6]  = mT.GetValue(3,1);
	
	//update Position
	x_dot->array[7]  = V.e[0];
	x_dot->array[8]  = V.e[1];
	x_dot->array[9]  = V.e[2];

	//update Velocity
	x_dot->array[10] = mF.GetValue(1,1);
	x_dot->array[11] = mF.GetValue(2,1);
	x_dot->array[12] = mF.GetValue(3,1);

	return;
}


//Assumes orientated wrt to axis and centered about O
int TestCuboidIntersection(C3dObject* pMesh, C3dObject* pCuboid, CContact* Return)
{
	int result=0;
	int i,j;

	for(i=0;i<sizeOfArray;i++)
		contactArray[i].bContactExists=false;

	CVector Max;
	Max.e[0]=pCuboid->vertList[0].x;
	Max.e[1]=pCuboid->vertList[0].y;
	Max.e[2]=pCuboid->vertList[0].z;

	for(i=0;i<3;i++)
	{
		if(Max.e[i]<0)
			Max.e[i]*=-1;
	}

	CMatrix oldO;
	CVector oldP;

	oldO = pMesh->data.Orientation;
	oldP = pMesh->data.Position;

	pMesh->data.Position    = pCuboid->data.Orientation.ReturnInverse()*(oldP-pCuboid->data.Position);
	pMesh->data.Orientation = pCuboid->data.Orientation.ReturnInverse()*oldO;
	
	for(i=0;i<pMesh->iNumOfVerts;i++)
	{
		int bounded;
		CVertex transformed;
		CVector v;
		transformed = pMesh->Transform(i);
		v.e[0] = transformed.x;
		v.e[1] = transformed.y;
		v.e[2] = transformed.z;

		Return[i].bContactExists = false;

		bounded=0;
		for(j=0;j<3;j++)
			if(v.e[j]<=Max.e[j] && v.e[j]>=-Max.e[j])
				bounded++;

		if(bounded==3)
		{
			//intersects
			result++;
			Return[i].bContactExists = true;
			Return[i].vertIndex      = i;
			
			int k,l;
			CVector temp;
			double MinDistance;
			double CurrDistance;

			//Find normal of closest plane
			for(k=0;k<3;k++)
				for(l=-1;l<=1;l+=2)
				{

					temp*=0;
					temp.e[k]=Max.e[k]*l;
					temp-=v;
					CurrDistance=temp.Length();
					if((k==0 && l==-1) || MinDistance>CurrDistance)
					{
						MinDistance = CurrDistance;
						Return[i].Normal*=0;
						Return[i].Normal.e[k] = l;
					}
				}
		}

		//Check Next Vertex
	}

	pMesh->data.Orientation=oldO;
	pMesh->data.Position=oldP;

	return result;
}



//Our main collision response routine
void UpdateMotionData(C3dObject* pMesh, int NoOfObjects)
{
	int i,j;
	double t = Global_T;
	CVector g;
	CVectorList vlStart, vlEnd;

	if(!vlStart.setDimension(22))
	{
		printf("*** Ran Out Of Memory ***\n");
		return;
	}

	if(!vlEnd.setDimension(22))
	{
		printf("*** Ran Out Of Memory ***\n");
		return;
	}

	g.e[0] = 0;
	g.e[1] = 0;
	g.e[2] = -9.8*Global_UnitsPerMetre;


	//Store current state as old state
	for(i=0;i<NoOfObjects;i++)
	{
		pMesh[i].data.old_AngularMomentum = pMesh[i].data.new_AngularMomentum;
		pMesh[i].data.old_Orientation     = pMesh[i].data.new_Orientation;
		pMesh[i].data.old_Position        = pMesh[i].data.new_Position;
		pMesh[i].data.old_Velocity        = pMesh[i].data.new_Velocity;

		pMesh[i].data.Position=pMesh[i].data.old_Position;
		QuaternionToMatrix(pMesh[i].data.old_Orientation,pMesh[i].data.Orientation);
	}

	
	//Resolve All Impulses

	//All collisions are assumed to be of the type vertex-face
	//with object A contributing a vertex, and B the normal
	int     Ai,   Bi;
	double  Am,   Bm;
	CVector Av,   Bv;
	CVector AL,   BL;
	CMatrix AInv, BInv;
	CVector Aw,   Bw;
	CVector Ar,   Br;
	CVector Arxn, Brxn;
	CVector n;


	//Object A is the object we are simulating
	Ai=1;

	//Object B is the floor
	Bi=0;

	for(Ai=1;Ai<NoOfObjects;Ai++)
	{
		//if(Bi!=0) continue;
		if(Ai==Bi)
			continue;

		//Update m,Iinv
		Am   = pMesh[Ai].data.Mass;
		AInv = pMesh[Ai].data.Orientation
			 * pMesh[Ai].data.InverseOfTheInertiaTensor
			 * pMesh[Ai].data.Orientation.ReturnTranspose();
		
		Bm   = pMesh[Bi].data.Mass;
		BInv = pMesh[Bi].data.Orientation
			 * pMesh[Bi].data.InverseOfTheInertiaTensor
			 * pMesh[Bi].data.Orientation.ReturnTranspose();

		//Check if testing two Infinite Masses colliding
		if(Am<0 && Bm<0)
			continue;

		//check if there is an intersection
		if(0!=TestCuboidIntersection(pMesh+Ai,pMesh+Bi,contactArray))
		{
			bool bApplyImpulse;

			bApplyImpulse = true;
			int county=0;
			while(bApplyImpulse)//everytime we apply an impulse we have to re-resolve all collisions
			{
				//Update v,L,w
				Av = pMesh[Ai].data.new_Velocity;
				AL = pMesh[Ai].data.new_AngularMomentum;
				Aw = AInv*AL;

				Bv = pMesh[Bi].data.new_Velocity;
				BL = pMesh[Bi].data.new_AngularMomentum;
				Bw = BInv*BL;

				bApplyImpulse = false;
				for(j=0;j<sizeOfArray;j++)
				if(contactArray[j].bContactExists)
				{
					//Update r,n
					Ar = pMesh[Ai].TransformToWorld(contactArray[j].vertIndex)-pMesh[Ai].data.new_Position;
					Br = pMesh[Ai].TransformToWorld(contactArray[j].vertIndex)-pMesh[Bi].data.new_Position;
					n  = pMesh[Bi].data.Orientation*contactArray[j].Normal;
					n.Normalize();

					//Check if separation velocity is +ve
					CVector ObjVelA;
					CVector ObjVelB;//cuboid

					ObjVelA = Av+Aw.Cross(Ar);
					ObjVelB = Bv+Bw.Cross(Br);

					double SepVelocity = (ObjVelA-ObjVelB).Dot(n);
					if(-0.0000001>SepVelocity)
					{
						//we have a collision and need to apply an impulse
						bApplyImpulse=true;
						break;
					}
				}

				//Apply Impulse at contact j
				if(bApplyImpulse)
				{
					double J;
					double e;
					
					e   = pMesh[Ai].data.CoefficientOfRestitution
						* pMesh[Bi].data.CoefficientOfRestitution;
					
					Arxn = Ar.Cross(n);
					Brxn = Br.Cross(n);
					
					double Am_reciprocal;
					double Bm_reciprocal;

					if(Am<=0)
						Am_reciprocal=0;
					else
						Am_reciprocal=1/Am;
					
					if(Bm<=0)
						Bm_reciprocal=0;
					else
						Bm_reciprocal=1/Bm;
					

					J = -(1+e) * (Av.Dot(n)+Arxn.Dot(Aw)-Bv.Dot(n)-Brxn.Dot(Bw))
						/ (Am_reciprocal*(1+Arxn.Dot(AInv*Arxn))
						+  Bm_reciprocal*(1+Brxn.Dot(BInv*Brxn)));
					
					
					pMesh[Ai].data.new_Velocity       +=Am_reciprocal*J*n;
					pMesh[Ai].data.new_AngularMomentum+=Am_reciprocal*J*Arxn;
					pMesh[Bi].data.new_Velocity       -=Bm_reciprocal*J*n;
					pMesh[Bi].data.new_AngularMomentum-=Bm_reciprocal*J*Brxn;
				}

			}//end of while

		}//end of check for intersection
	}



	//Integrate motion forward
	for(i=1;i<NoOfObjects;i++)
	{
		//Check for small Velocities

		TestCuboidIntersection(pMesh+i,pMesh,contactArray);

		//Update v,L,w,Iinv
		CVector v   = pMesh[i].data.new_Velocity;
		CVector L   = pMesh[i].data.new_AngularMomentum;
		CMatrix Inv = pMesh[i].data.Orientation
					* pMesh[i].data.InverseOfTheInertiaTensor
					* pMesh[i].data.Orientation.ReturnTranspose();
		CVector w   = Inv*L;

		for(j=0;j<sizeOfArray;j++)
		{
			if(contactArray[j].bContactExists)
			{
				//Update r,n
				CVector r = pMesh[i].TransformToWorld(contactArray[j].vertIndex)-pMesh[i].data.new_Position;
				CVector n = contactArray[j].Normal;
				n.Normalize();
				
				//Check if separation velocity is +ve
				CVector ObjVelA;
				CVector ObjVelB;//cuboid
				
				ObjVelA = v+w.Cross(r);
				ObjVelB = CVector(0,0,0);
				
				double SepVelocity = (ObjVelA-ObjVelB).Dot(n);

				//Identify the resting contacts
				if(Global_VelocityZero<SepVelocity)
					contactArray[j].bContactExists=false;
			}
		}
		//End of Check for small Velocities

		//Convert the state of the object into a a form we can integrate using ode()
		StateToArray(
			vlStart.array,
			pMesh[i].data.new_Orientation,
			pMesh[i].data.new_AngularMomentum,
			pMesh[i].data.new_Position,
			pMesh[i].data.new_Velocity,
			pMesh[i].data.InverseOfTheInertiaTensor);
		
		Global_pMesh = pMesh+i;

		//Numerically integrate the state forward in time
		//The equations of motion are given by the function qDash()
		ode(&vlStart,&vlEnd,0,t,1,qDash);
		
		//Update the state of the object
		CMatrix ignore;
		ArrayToState(
			vlEnd.array,
			pMesh[i].data.new_Orientation,
			pMesh[i].data.new_AngularMomentum,
			pMesh[i].data.new_Position,
			pMesh[i].data.new_Velocity,
			ignore);
	

		pMesh[i].data.Position=pMesh[i].data.new_Position;
		QuaternionToMatrix(pMesh[i].data.new_Orientation,pMesh[i].data.Orientation);
	}

	return;
}

